# -*- coding: utf-8 -*-
"""
Utilities for handling the reparameterisations.
"""

from .base import Reparameterisation
from .null import NullReparameterisation


def get_reparameterisation(reparameterisation, defaults=None):
    """Function to get a reparameterisation class from a name

    Parameters
    ----------
    reparameterisation : str, \
            :obj:`nessai.reparameterisations.Reparameterisation`
        Name of the reparameterisations to return or a class that inherits from
        :obj:`~nessai.reparameterisations.Reparameterisation`
    defaults : dict, optional
        Dictionary of known reparameterisations that overrides the defaults.

    Returns
    -------
    :obj:`nessai.reparameteristaions.Reparameterisation`
        Reparameterisation class.
    dict
        Keyword arguments for the specific reparameterisation.
    """
    if defaults is None:
        from . import default_reparameterisations

        defaults = default_reparameterisations
    if isinstance(reparameterisation, str):
        rc, kwargs = defaults.get(reparameterisation, (None, None))
        if rc is None:
            raise ValueError(
                f"Unknown reparameterisation: {reparameterisation}"
            )
        else:
            if kwargs is None:
                kwargs = {}
            else:
                kwargs = kwargs.copy()
            return rc, kwargs
    elif reparameterisation is None:
        return NullReparameterisation, {}
    elif isinstance(reparameterisation, type) and issubclass(
        reparameterisation, Reparameterisation
    ):
        return reparameterisation, {}
    else:
        raise TypeError(
            "Reparameterisation must be a str, None, or class that "
            "inherits from `Reparameterisation`"
        )



# -*- coding: utf-8 -*-
"""
Utilities related to drawing samples.
"""
import numpy as np
from scipy import stats
from scipy.special import gammaincinv


def compute_radius(n, q=0.95):
    """Compute the radius that contains a fraction of the total probability \
         in an n-dimensional unit Gaussian.

    Uses the inverse CDF of a chi-distribution with n degrees of freedom.

    Parameters
    ----------
    n : int
        Number of dimensions
    q : float
        Fraction of the total probability

    Returns
    -------
    float
        Radius
    """
    return stats.chi.ppf(q, n)


def draw_surface_nsphere(dims, r=1, N=1000):
    """
    Draw N points uniformly from  n-1 sphere of radius r using Marsaglia's
    algorithm. E.g for 3 dimensions returns points on a 'regular' sphere.

    See Marsaglia (1972)

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-sphere, if specified it is used to rescale the samples
    N : int, optional
        Number of samples to draw

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    x = np.random.randn(N, dims)
    R = np.sqrt(np.sum(x**2.0, axis=1))[:, np.newaxis]
    z = x / R
    return r * z


def draw_nsphere(dims, r=1, N=1000, fuzz=1.0):
    """
    Draw N points uniformly within an n-sphere of radius r

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball
    N : int, optional
        Number of samples to draw
    fuzz : float, optional
        Fuzz factor by which to increase the radius of the n-ball

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    x = draw_surface_nsphere(dims, r=1, N=N)
    R = np.random.uniform(0, 1, (N, 1))
    z = R ** (1 / dims) * x
    return fuzz * r * z


def draw_uniform(dims, r=(1,), N=1000, fuzz=1.0):
    """
    Draw from a uniform distribution on [0, 1].

    Deals with extra input parameters used by other draw functions

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball. (Ignored by this function)
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the n-ball. (Ignored by
        this function)

    Returns
    -------
    ndarraay
        Array of samples with shape (N, dims)
    """
    return np.random.uniform(0, 1, (N, dims))


def draw_gaussian(dims, r=1, N=1000, fuzz=1.0):
    """
    Wrapper for numpy.random.randn that deals with extra input parameters
    r and fuzz

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the n-ball

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    return np.random.randn(N, dims)


def draw_truncated_gaussian(dims, r, N=1000, fuzz=1.0, var=1):
    """
    Draw N points from a truncated gaussian with a given a radius

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float
        Radius of the truncated Gaussian
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the truncated Gaussian

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    sigma = np.sqrt(var)
    u_max = stats.chi.cdf(r * fuzz / sigma, df=dims)
    u = np.random.uniform(0, u_max, N)
    p = sigma * stats.chi.ppf(u, df=dims)
    x = np.random.randn(p.size, dims)
    points = (p * x.T / np.sqrt(np.sum(x**2.0, axis=1))).T
    return points


class NDimensionalTruncatedGaussian:
    """Class for sampling from a radially truncated n-dimensional Gaussian

    Parameters
    ----------
    dims :
        The number of dimensions
    radius :
        The radius for the truncation
    fuzz : float
        The fuzz factor
    """

    def __init__(
        self,
        dims: int,
        radius: float,
        fuzz: float = 1.0,
    ) -> None:
        self.dims = dims
        self.radius = radius
        self.fuzz = fuzz
        self.chi = stats.chi(df=self.dims)

        self.u_max = self.chi.cdf(self.radius * self.fuzz)

    def sample(self, N: int) -> np.ndarray:
        """Sample from the distribution.

        Parameters
        ----------
        n : int
            Number of samples to draw

        Returns
        -------
        numpy.ndarray
            Array of samples of shape [n, dims].
        """
        u = self.u_max * np.random.rand(N)
        # Inverse CDF of a chi-distribution
        p = np.sqrt(2 * gammaincinv(0.5 * self.dims, u))
        x = np.random.randn(self.dims, N)
        points = (p * x / np.sqrt(np.sum(x**2.0, axis=0))).T
        return points



# -*- coding: utf-8 -*-
"""
Test utilities for computing distances.
"""
import numpy as np
import pytest
from unittest.mock import patch

from nessai.utils.distance import compute_minimum_distances


def test_minimum_distance():
    """Test the minimum distance function"""
    samples = np.array([[1], [2], [4]])
    d = np.array([[0, 1, 3], [1, 0, 2], [3, 2, 0]], dtype=float)
    with patch("nessai.utils.distance.distance.cdist", return_value=d) as mock:
        dmin = compute_minimum_distances(samples, metric="test")

    mock.assert_called_once_with(samples, samples, "test")
    np.testing.assert_array_equal(dmin, np.array([1, 1, 2]))


@pytest.mark.integration_test
@pytest.mark.parametrize("metric", ["euclidean", "minkowski"])
def test_minimum_distance_integration(metric):
    """Integration test for the minimum distance"""
    samples = np.array([[1], [2], [4]])
    dmin = compute_minimum_distances(samples, metric=metric)
    np.testing.assert_array_equal(dmin, np.array([1, 1, 2]))



# -*- coding: utf-8 -*-
"""
Tests for rescaling functions
"""
import multiprocessing
from multiprocessing.dummy import Pool
import numpy as np
import pytest
import sys
from unittest.mock import MagicMock, patch

from nessai.utils.multiprocessing import (
    batch_evaluate_function,
    check_multiprocessing_start_method,
    check_vectorised_function,
    initialise_pool_variables,
    get_n_pool,
    log_likelihood_wrapper,
)


def test_pool_variables():
    """Assert initialising the variables and calling the likelihood work"""
    model = MagicMock()
    model.log_likelihood = lambda x: x
    initialise_pool_variables(model)
    pool = Pool(1)
    out = pool.map(log_likelihood_wrapper, [1, 2, 3])
    pool.close()
    pool.terminate()
    assert out == [1, 2, 3]

    # Reset to the default value
    initialise_pool_variables(None)


@pytest.mark.parametrize("method", ["fork", "forkserver", "spawn"])
def test_check_multiprocessing_start_method(method, caplog):
    """
    Test check multiprocessing start method passes for all methods
    """
    with patch("multiprocessing.get_start_method", return_value=method):
        check_multiprocessing_start_method()
    if method != "fork":
        assert "This may lead to high memory usage or errors" in caplog.text


@pytest.mark.integration_test
@pytest.mark.skip_on_windows
def test_check_multiprocessing_start_method_integration():
    """Integration test for checking the start method."""
    mp = multiprocessing.get_context("fork")
    with patch("multiprocessing.get_start_method", mp.get_start_method):
        check_multiprocessing_start_method()


@pytest.mark.parametrize("method", ["fork", "forkserver", "spawn"])
@pytest.mark.integration_test
def test_check_multiprocessing_start_method_error_integration(method, caplog):
    """Integration test for checking the start method prints a warning"""
    if sys.platform == "win32" and method != "spawn":
        pytest.skip("Windows only supports the 'spawn' start method")
    mp = multiprocessing.get_context(method)
    with patch("multiprocessing.get_start_method", mp.get_start_method):
        check_multiprocessing_start_method()
    if method != "fork":
        assert "This may lead to high memory usage or errors" in caplog.text


def test_model_error():
    """Assert an error is raised in the global variables have not been \
        initialised
    """
    model = MagicMock()
    model.log_likelihood = lambda x: x
    pool = Pool(1)
    with pytest.raises(AttributeError) as excinfo:
        pool.map(log_likelihood_wrapper, [1, 2, 3])
    assert "'NoneType' object has no attribute 'log_likelihood'" in str(
        excinfo.value
    )
    pool.close()
    pool.terminate()


def test_get_n_pool_processes():
    """ "Assert the correct value is returned if the pool has a `_processes`
    attribute.
    """
    pool = MagicMock(spec=Pool)
    pool._processes = 4
    assert get_n_pool(pool) == 4


def test_get_n_pool_ray():
    """ "Assert the correct value is returned if the pool has a `_actor_pool`
    attribute (e.g. ray.util.multiprocessing.Pool).
    """
    pool = MagicMock()
    del pool._processes
    pool._actor_pool = 4 * [0]
    assert get_n_pool(pool) == 4


def test_get_n_pool_unknown():
    """ "Assert None is returned if the type of pool is not known."""
    pool = MagicMock()
    del pool._processes
    del pool._actor_pool
    assert get_n_pool(pool) is None


@pytest.mark.parametrize(
    "n, vectorised, chunksize, expected_calls",
    [
        (10, True, None, 1),
        (10, False, None, 10),
        (10, True, 4, 3),
        (10, False, 4, 10),
    ],
)
def test_batch_evaluate_function_vectorised(
    n,
    vectorised,
    chunksize,
    expected_calls,
):

    global ncalls
    ncalls = 0

    def func(x):
        global ncalls
        ncalls = ncalls + 1
        return (-(x**2)).sum(-1)

    x = np.random.randn(n, 2)

    out = batch_evaluate_function(
        func, x, vectorised=vectorised, chunksize=chunksize
    )
    assert out.shape == (10,)
    assert ncalls == expected_calls


def test_check_vectorised_function_pass():
    def func(x):
        return (-(x**2)).sum(axis=-1)

    x = np.arange(10)[:, np.newaxis]

    out = check_vectorised_function(func, x, dtype="float64")
    assert out is True


def test_check_vectorised_function_error():
    def func(x):
        len(x)
        return 0.5

    x = np.arange(10)[:, np.newaxis]

    out = check_vectorised_function(func, x)
    assert out is False


def test_check_vectorised_function_invalid_output():
    def func(x):
        return 0.5

    x = np.arange(10)

    out = check_vectorised_function(func, x)
    assert out is False


def test_check_vectorised_function_invalid_value():
    def func(x):
        return np.cumsum(x)

    x = np.arange(10)

    out = check_vectorised_function(func, x)
    assert out is False


def test_check_vectorised_function_invalid_input():
    with pytest.raises(ValueError, match=r"Input has length <= 1"):
        check_vectorised_function(lambda x: x, np.array([0]))



# -*- coding: utf-8 -*-
"""
Utilities related to optimisation.
"""
import logging
from typing import Optional

import numpy as np
from scipy.optimize import minimize
from scipy.special import logsumexp


logger = logging.getLogger(__name__)


def optimise_meta_proposal_weights(
    samples: np.ndarray,
    log_q: np.ndarray,
    method: str = "SLSQP",
    options: Optional[dict] = None,
    initial_weights: Optional[np.ndarray] = None,
) -> np.ndarray:
    """Optimise the weights of the meta proposal.

    Uses :code:`scipy.optimize.minimize`.

    Parameters
    ----------
    samples
        Samples drawn from the initial meta proposal.
    log_q
        Array of log probabilities for each proposal for each sample.
    method
        Optimisation method to use. See scipy docs for details.
    options
        Dictionary of options for :code:`scipy.optimize.minimize`.
    """
    if options is None and method == "SLSQP":
        options = dict(ftol=1e-10)

    n_prop = log_q.shape[-1]
    counts = np.unique(samples["it"], return_counts=True)[1]
    if initial_weights is None:
        initial_weights = counts / counts.sum()
    else:
        initial_weights = initial_weights / initial_weights.sum()

    log_Z = logsumexp(samples["logL"] - samples["logQ"]) - np.log(len(samples))

    log_pr = samples["logL"] - log_Z
    log_pr -= logsumexp(log_pr)
    pr = np.exp(log_pr)

    def loss(weights):
        """Computes the KL"""
        log_Q = logsumexp(log_q, b=weights, axis=1)
        return -np.mean(pr * log_Q)

    # Weights must sum to one
    constraint = {"type": "eq", "fun": lambda x: 1 - x.sum()}

    logger.info("Starting optimisation")
    result = minimize(
        loss,
        initial_weights,
        constraints=constraint,
        bounds=n_prop * [(0, 1)],
        method=method,
        options=options,
    )
    logger.info("Finished optimisation")
    logger.debug(f"Final weights: {result.x}")

    return np.array(result.x)



#!/usr/bin/env python

# Example of using nessai with the eggbox function. Nessai is not well
# suited to this sorted problem. The augmented proposal method is used to
# to improve the results but ultimately likelihoods that this multi-modal
# are a challenge.

import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger


output = "./outdir/eggbox/"
logger = setup_logger(output=output)


class EggboxModel(Model):
    """
    Eggbox problem from https://arxiv.org/pdf/0809.3437v1.pdf.

    Based on the example in cpnest: \
        https://github.com/johnveitch/cpnest/blob/master/examples/eggbox.py
    """

    def __init__(self, dims):
        self.names = [str(i) for i in range(dims)]
        self.bounds = {n: [0, 10 * np.pi] for n in self.names}

    def log_prior(self, x):
        """Log-prior."""
        log_p = np.log(self.in_bounds(x), dtype="float")
        for bounds in self.bounds.values():
            log_p -= np.log(bounds[1] - bounds[0])
        return log_p

    def log_likelihood(self, x):
        """Log-likelihood."""
        return (
            np.prod(np.cos(self.unstructured_view(x) / 2.0), axis=-1) + 2
        ) ** 5.0


flow_config = dict(
    patience=50,  # Make sure the flow trains for longer
    model_config=dict(n_blocks=6, n_neurons=8),
)

# Sampling the Eggbox will be inefficient, so we increase the maximum size of
# the pool of proposal points. We also allow the initial sampling without
# the flow to run until it becomes inefficient rather than a fixed iteration.
fs = FlowSampler(
    EggboxModel(2),
    output=output,
    flow_config=flow_config,
    resume=False,
    seed=1234,
    flow_class="AugmentedFlowProposal",
    augment_dims=2,  # Add two augment parameters
    maximum_uninformed="inf",  # Run initial sampling until it's inefficient
    max_poolsize_scale=50,  # Increase the maximum poolsize
    max_threads=3,
    n_pool=2,
)

fs.run()



# -*- coding: utf-8 -*-
"""
Utilities for configuring FlowModel.
"""
from .config import DEFAULT_MODEL_CONFIG, DEFAULT_FLOW_CONFIG
from ..flows.utils import get_n_neurons


def update_model_config(d):
    """Update the model (flow) configuration dictionary based on the defaults.

    Parameters
    ----------
    d : Union[dict, None]
        Dictionary with the current configuration. If None, then the default is
        used.

    Returns
    -------
    dict
        Updated configuration dictionary.

    Raises
    ------
    TypeError
        Raised if the input is not a dictionary or None.
    """
    default = DEFAULT_MODEL_CONFIG.copy()
    if d is None:
        return default
    elif not isinstance(d, dict):
        raise TypeError(
            "Must pass a dictionary to update the default model config"
        )
    else:
        default.update(d)
        default["n_neurons"] = get_n_neurons(
            n_neurons=default.get("n_neurons"),
            n_inputs=default.get("n_inputs"),
        )
    return default


def update_config(d):
    """
    Update the configuration dictionary to include the defaults.

    Notes
    -----
    The default configuration is specified in :py:mod:`nessai.flowmodel.config`


    The kwargs can contain any additional keyword arguments that are specific
    to the type of flow being used.

    Parameters
    ----------
    d : dict
        Dictionary with configuration

    Returns
    -------
    dict
        Dictionary with updated default configuration
    """
    default = DEFAULT_FLOW_CONFIG.copy()

    if d is None:
        default["model_config"] = update_model_config(None)
    elif not isinstance(d, dict):
        raise TypeError(
            "Must pass a dictionary to update the default " "trainer settings"
        )
    else:
        default.update(d)
        default["model_config"] = update_model_config(
            default.get("model_config", None)
        )

    if default["noise_type"] is not None and default["noise_scale"] is None:
        raise RuntimeError(
            "`noise_scale` must be specified when `noise_type` is given."
        )
    if isinstance(default["noise_scale"], float):
        if default["noise_type"] is None:
            default["noise_type"] = "constant"
    elif default["noise_scale"] is not None:
        raise TypeError(
            "`noise_scale` must be a float. "
            f"'Got type: {type(default['noise_scale'])}"
        )

    return default



# -*- coding: utf-8 -*-
"""
Tests for modules/functions that are soon to be deprecated.
"""
import pytest


def test_nested_sampler_deprecation():
    """Assert a warning is raised with nessai.nestedsampler is imported."""
    with pytest.warns(FutureWarning) as record:
        from nessai import nestedsampler  # noqa
    assert "`nessai.nestedsampler` is deprecated" in str(record[0].message)


def test_lulinear_warning():
    """Assert a warning is raised when LULinear is imported"""
    with pytest.warns(FutureWarning) as record:
        from nessai.flows.transforms import LULinear  # noqa
    assert "`nessai.flows.transforms.LULinear` is deprecated" in str(
        record[0].message
    )


def test_bilbyutils_warning():
    """Assert a warning is raised if bilbyutils is imported"""
    with pytest.warns(
        FutureWarning, match=r"`nessai.utils.bilbyutils` is deprecated"
    ):
        from nessai.utils.bilbyutils import get_all_kwargs  # noqa



#!/usr/bin/env python

# Example of using AugmentedFlowProposal for a multimodal model
# Note that this feature is experimental and not fully tested.

import numpy as np
from scipy.stats import norm

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger

output = "./outdir/augmented_example/"
logger = setup_logger(output=output)


class MultimodalModel(Model):
    """A multimodal gaussian model."""

    def __init__(self, dims=2):
        self.names = [str(i) for i in range(dims)]
        self.bounds = {i: [-10, 10] for i in self.names}

    def log_prior(self, x):
        """Log-prior"""
        log_p = np.log(self.in_bounds(x), dtype="float")
        for bounds in self.bounds.values():
            log_p -= np.log(bounds[1] - bounds[0])
        return log_p

    def log_likelihood(self, x):
        """Log likelihood."""
        log_l = np.zeros(x.size)
        for n in self.names:
            log_l += np.log(norm(-5).pdf(x[n]) + norm(5).pdf(x[n]))
        return log_l


# We add `augment_dims` which help the sampler bridge disconnected regions of
# probability. There additional parameters are Gaussian and included in the
# parameters transformed by the flow.
fs = FlowSampler(
    MultimodalModel(4),
    output=output,
    resume=False,
    seed=1234,
    augment_dims=2,
    flow_class="AugmentedFlowProposal",
)

# And go!
fs.run()



# -*- coding: utf-8 -*-
"""
Proposal methods for used in the nested sampler.
"""
from .analytic import AnalyticProposal
from .augmented import AugmentedFlowProposal
from .flowproposal import FlowProposal
from .rejection import RejectionProposal

__all__ = [
    "AnalyticProposal",
    "AugmentedFlowProposal",
    "FlowProposal",
    "RejectionProposal",
]



# -*- coding: utf-8 -*-
"""
Tests for `nessai.model`
"""
import datetime
import logging
import numpy as np
import pytest
from scipy.stats import norm
from unittest.mock import MagicMock, call, create_autospec, patch

from nessai import config
from nessai.livepoint import numpy_array_to_live_points
from nessai.model import Model, OneDimensionalModelError
from nessai.utils.multiprocessing import (
    initialise_pool_variables,
    log_likelihood_wrapper,
)
from nessai.utils.testing import (
    assert_structured_arrays_equal,
)


class EmptyModel(Model):
    def log_prior(self, x):
        return None

    def log_likelihood(self, x):
        return None


class BasicModel(Model):
    def __init__(self):
        self.bounds = {"x": [-5, 5], "y": [-5, 5]}
        self.names = ["x", "y"]

    def log_prior(self, x):
        log_p = np.log(self.in_bounds(x), dtype="float")
        for n in self.names:
            log_p -= np.log(self.bounds[n][1] - self.bounds[n][0])
        return log_p

    def log_likelihood(self, x):
        log_l = np.ones(x.size)
        for pn in self.names:
            log_l += norm.logpdf(x[pn])
        return log_l


@pytest.fixture
def model():
    return create_autospec(Model, _pool_configured=False)


@pytest.fixture
def live_point(integration_model):
    return integration_model.new_point()


@pytest.fixture
def live_points(integration_model):
    return integration_model.new_point(10)


def test_names(model):
    """Assert names returns the correct value."""
    model._names = ["x", "y"]
    assert Model.names.__get__(model) is model._names


def test_names_setter(model):
    """Assert names is correctly set"""
    Model.names.__set__(model, ["x", "y"])
    assert model._names == ["x", "y"]


def test_names_invalid_type(model):
    """Assert an error is raised if `names` is not a list."""
    with pytest.raises(TypeError) as excinfo:
        Model.names.__set__(model, True)
    assert "`names` must be a list" in str(excinfo.value)


def test_names_empty_list(model):
    """Assert an error is raised if `names` is empty."""
    with pytest.raises(ValueError) as excinfo:
        Model.names.__set__(model, [])
    assert "`names` list is empty" in str(excinfo.value)


def test_names_1d_list(model):
    """Assert an error is raised if `names` is 1d"""
    with pytest.raises(OneDimensionalModelError) as excinfo:
        Model.names.__set__(model, ["x"])
    assert "names list has length 1" in str(excinfo.value)


def test_bounds(model):
    """Assert bounds returns the correct value."""
    model._bounds = {"x": [-1, 1], "y": [-1, 1]}
    assert Model.bounds.__get__(model) is model._bounds


def test_bounds_setter(model):
    """Assert bounds is correctly set"""
    Model.bounds.__set__(model, {"x": [-1, 1], "y": [-2, 2]})
    assert list(model._bounds.keys()) == ["x", "y"]
    np.testing.assert_array_equal(model._bounds["x"], [-1, 1])
    np.testing.assert_array_equal(model._bounds["y"], [-2, 2])


def test_bounds_invalid_type(model):
    """Assert an error is raised if `bounds` is not a dictionary."""
    with pytest.raises(TypeError) as excinfo:
        Model.bounds.__set__(model, True)
    assert "`bounds` must be a dictionary" in str(excinfo.value)


def test_bounds_1d(model):
    """Assert an error is raised if `bounds` is 1d"""
    with pytest.raises(OneDimensionalModelError) as excinfo:
        Model.bounds.__set__(model, {"x": [0, 1]})
    assert "bounds dictionary has length 1" in str(excinfo.value)


@pytest.mark.parametrize("b", [[1], [1, 2, 3]])
def test_bounds_incorrect_length(model, b):
    """Assert an error is raised if `bounds` is 1d"""
    with pytest.raises(ValueError) as excinfo:
        Model.bounds.__set__(model, {"x": b, "y": [0, 1]})
    assert "Each entry in `bounds` must have length 2" in str(excinfo.value)


def test_dims(model):
    """Ensure dims are correct"""
    model.names = ["x", "y"]
    assert Model.dims.__get__(model) == 2


def test_dims_no_names(model):
    """Test the behaviour dims when names is empty"""
    model.names = []
    assert Model.dims.__get__(model) is None


def test_set_upper_lower(model):
    """Assert the upper and lower bounds are set correctly."""
    model.names = ["y", "x"]
    model.bounds = {"x": [0, 1], "y": [-1, 2]}
    Model._set_upper_lower(model)
    np.testing.assert_array_equal(model._lower, np.array([-1, 0]))
    np.testing.assert_array_equal(model._upper, np.array([2, 1]))


def test_lower_bounds(model):
    """Check the lower bounds are correctly set"""

    def func():
        model._lower = np.array([-1, -1])

    model._set_upper_lower = MagicMock(side_effect=func)
    model._lower = None
    bounds = Model.lower_bounds.__get__(model)
    model._set_upper_lower.assert_called_once()
    np.testing.assert_array_equal(bounds, np.array([-1, -1]))


def test_upper_bounds(model):
    """Check the upper bounds are correctly set"""

    def func():
        model._upper = np.array([1.0, 1.0])

    model._set_upper_lower = MagicMock(side_effect=func)
    model._upper = None
    bounds = Model.upper_bounds.__get__(model)
    model._set_upper_lower.assert_called_once()
    np.testing.assert_array_equal(bounds, np.array([1, 1]))


@pytest.mark.parametrize("value", [True, False])
def test_vectorised_likelihood(model, value):
    """Assert the correct value is stored if allow_vectorised is True"""
    model._vectorised_likelihood = None
    model.allow_vectorised = True
    model.new_point = MagicMock(
        side_effect=[
            np.random.rand(1).astype([("x", "f8")]) for _ in range(10)
        ]
    )

    with patch("nessai.model.check_vectorised_function", return_value=value):
        out = Model.vectorised_likelihood.__get__(model)

    assert model._vectorised_likelihood is value
    assert out is value


@pytest.mark.parametrize("error", [TypeError, ValueError])
def test_vectorised_likelihood_not_vectorised_error(model, error):
    """
    Assert the value is False if the likelihood is not vectorised and raises
    an error.
    """

    def dummy_likelihood(x):
        if hasattr(x, "__len__"):
            raise error
        else:
            return np.log(np.random.rand())

    model._vectorised_likelihood = None
    model.log_likelihood = MagicMock(side_effect=dummy_likelihood)
    model.new_point = MagicMock(return_value=np.random.rand(10))

    out = Model.vectorised_likelihood.__get__(model)
    assert model._vectorised_likelihood is False
    assert out is False


def test_vectorised_likelihood_allow_vectorised_false(model):
    """Assert vectorised_likelihood is False if allow_vectorised is False"""
    model.allow_vectorised = False
    model.log_likelihood = MagicMock()
    model._vectorised_likelihood = None
    out = Model.vectorised_likelihood.__get__(model)
    model.log_likelihood.assert_not_called()
    assert out is False


def test_vectorised_likelihood_setter(model):
    """Assert the setter sets the correct variable."""
    Model.vectorised_likelihood.__set__(model, "test")
    assert model._vectorised_likelihood == "test"


@pytest.mark.parametrize("check_value", [True, False])
@pytest.mark.parametrize("allow_vectorised", [True, False])
def test_vectorised_log_prior(model, check_value, allow_vectorised):

    model._vectorised_prior = None
    model.allow_vectorised_prior = allow_vectorised

    with patch(
        "nessai.model.check_vectorised_function", return_value=check_value
    ):
        out = Model.vectorised_prior.__get__(model)

    assert out is (check_value and allow_vectorised)
    assert model._vectorised_prior is (check_value and allow_vectorised)


def test_vectorised_prior_setter(model):
    """Assert the setter sets the correct value"""
    Model.vectorised_prior.__set__(model, "test")
    assert model._vectorised_prior == "test"


def test_in_bounds(model):
    """Test the `in_bounds` method.

    Tests both finite and infinite prior bounds.
    """
    x = numpy_array_to_live_points(
        np.array([[0.5, 1.0], [2.0, 1.0]]), ["x", "y"]
    )
    model.names = ["x", "y"]
    model.bounds = {"x": [0, 1], "y": [-np.inf, np.inf]}
    val = Model.in_bounds(model, x)
    np.testing.assert_array_equal(val, np.array([True, False]))


def test_parameter_in_bounds(model):
    """Test parameter in bounds method."""
    x = np.array([0, 0.5, 1, 3])
    model.names = ["x", "y"]
    model.bounds = {"x": [0, 1], "y": [0, 4]}
    val = Model.parameter_in_bounds(model, x, "x")
    np.testing.assert_array_equal(val, np.array([True, True, True, False]))


def test_sample_parameter(model):
    """Assert an error is raised."""
    with pytest.raises(NotImplementedError) as excinfo:
        Model.sample_parameter(model, "x", n=2)
    assert "User must implement this method!" in str(excinfo.value)


def test_new_point_single(model):
    """Test the new point when asking for 1 point"""
    model._single_new_point = MagicMock()
    Model.new_point(model, N=1)
    model._single_new_point.assert_called_once()


def test_single_new_point(model):
    """Test the method that draw one point within the prior bounds"""
    model.names = ["x", "y"]
    model.bounds = {"x": [-1, 1], "y": [-2, 2]}
    model.lower_bounds = np.array([-1, -2])
    model.upper_bounds = np.array([1, 2])
    model.log_prior = MagicMock(return_value=0)
    model.dims = 2
    x = Model._single_new_point(model)
    assert (x["x"] >= -1) & (x["x"] <= 1)
    assert (x["y"] >= -2) & (x["y"] <= 2)


def test_new_point_multiple(model):
    """Test the new point when asking for multiple points"""
    model._multiple_new_points = MagicMock()
    Model.new_point(model, N=10)
    model._multiple_new_points.assert_called_once_with(10)


def test_multiple_new_points(model):
    """Test the method that draws multiple points within the prior bounds"""
    n = 10
    model.names = ["x", "y"]
    model.bounds = {"x": [-1, 1], "y": [-2, 2]}
    model.lower_bounds = np.array([-1, -2])
    model.upper_bounds = np.array([1, 2])
    model.log_prior = MagicMock(return_value=np.zeros(10))
    model.dims = 2
    x = Model._multiple_new_points(model, N=n)
    assert x.size == n
    assert ((x["x"] >= -1) & (x["x"] <= 1)).all()
    assert ((x["y"] >= -2) & (x["y"] <= 2)).all()


def test_multiple_new_points_reject(model):
    """Assert the correct number of points are returned in some are rejected"""
    n = 3
    model.names = ["x", "y"]
    model.bounds = {"x": [-1, 1], "y": [-2, 2]}
    model.lower_bounds = np.array([-1, -2])
    model.upper_bounds = np.array([1, 2])
    model.log_prior = MagicMock(
        side_effect=2 * [np.array([-np.inf, 0.0, 0.0])]
    )
    model.dims = 2
    x = Model._multiple_new_points(model, N=n)
    assert x.size == n
    assert ((x["x"] >= -1) & (x["x"] <= 1)).all()
    assert ((x["y"] >= -2) & (x["y"] <= 2)).all()


def test_new_point_log_prob(model):
    """Test the log prob for new points.

    Should be zero.
    """
    x = numpy_array_to_live_points(np.random.randn(2, 1), ["x"])
    log_prob = Model.new_point_log_prob(model, x)
    assert log_prob.size == 2
    assert (log_prob == 0).all()


@pytest.mark.integration_test
def test_new_point_integration(integration_model):
    """
    Test the default method for generating a new point with the bounds.

    Uses the model defined in `conftest.py` with bounds [-5, 5] for
    x and y.
    """
    new_point = integration_model.new_point()
    log_q = integration_model.new_point_log_prob(new_point)
    assert (new_point["x_0"] < 5) & (new_point["x_0"] > -5)
    assert (new_point["x_1"] < 5) & (new_point["x_1"] > -5)
    assert log_q == 0


@pytest.mark.integration_test
def test_new_point_multiple_integration(integration_model):
    """
    Test drawing multiple new points from the model

    Uses the model defined in `conftest.py` with bounds [-5, 5] for
    x_0 and x_1.
    """
    new_points = integration_model.new_point(N=100)
    log_q = integration_model.new_point_log_prob(new_points)
    assert new_points.size == 100
    assert all(np.isnan(new_points["logP"]))
    assert all(new_points["x_0"] < 5) & all(new_points["x_1"] > -5)
    assert all(new_points["x_1"] < 5) & all(new_points["x_0"] > -5)
    assert (log_q == 0).all()


def test_likelihood_evaluations(model, live_point):
    """
    Test `evaluate_log_likelihood` and ensure the counter increases.
    """
    model.likelihood_evaluations = 1
    model.log_likelihood = MagicMock(return_value=2)
    log_l = Model.evaluate_log_likelihood(model, live_point)

    model.log_likelihood.assert_called_once_with(live_point)
    assert log_l == 2
    assert model.likelihood_evaluations == 2


def test_likelihood_evaluations_vectorised(model, live_points):
    """
    Test `evaluate_log_likelihood` and ensure the counter increases.
    """
    out = np.random.randn(live_points.size)
    model.likelihood_evaluations = 1
    model.log_likelihood = MagicMock(return_value=out)
    log_l = Model.evaluate_log_likelihood(model, live_points)

    model.log_likelihood.assert_called_once_with(live_points)
    assert log_l is out
    assert model.likelihood_evaluations == (1 + live_points.size)


def test_log_prior(model):
    """Verify the log prior raises a NotImplementedError"""
    with pytest.raises(NotImplementedError):
        Model.log_prior(model, 1)


def test_log_likelihood(model):
    """Verify the log likelihood raises a NotImplementedError"""
    with pytest.raises(NotImplementedError):
        Model.log_likelihood(model, 1)


def test_to_unit_hypercube(model):
    """Assert an error is raised by default"""
    with pytest.raises(NotImplementedError):
        Model.to_unit_hypercube(model, 1)


def test_from_unit_hypercube(model):
    """Assert an error is raised by default"""
    with pytest.raises(NotImplementedError):
        Model.from_unit_hypercube(model, 1)


def test_missing_log_prior():
    """
    Test to ensure a model cannot be instantiated without a log-prior method.
    """

    class TestModel(Model):
        def __init__(self):
            self.bounds = {"x": [-5, 5], "y": [-5, 5]}
            self.names = ["x", "y"]

        def log_likelihood(self, x):
            return x

    with pytest.raises(TypeError) as excinfo:
        TestModel()
    assert "Can't instantiate abstract class TestModel" in str(excinfo.value)


def test_missing_log_likelihood():
    """
    Test to ensure a model cannot be instantiated without a log-likelihood
    method.
    """

    class TestModel(Model):
        def __init__(self):
            self.bounds = {"x": [-5, 5], "y": [-5, 5]}
            self.names = ["x", "y"]

        def log_prior(self, x):
            return 0.0

    with pytest.raises(TypeError) as excinfo:
        TestModel()
    assert "Can't instantiate abstract class TestModel" in str(excinfo.value)


def test_model_1d_error():
    """Assert an error is raised if the model is one dimensional."""

    class TestModel(EmptyModel):
        def __init__(self):
            self.names = ["x"]
            self.bounds = {"x": [-5, 5]}

    with pytest.raises(OneDimensionalModelError) as excinfo:
        TestModel()
    assert "nessai is not designed to handle one-dimensional models" in str(
        excinfo.value
    )


def test_verify_new_point():
    """
    Test `Model.verify_model` and ensure a model with an ill-defined
    prior function raises the correct error
    """

    class BrokenModel(BasicModel):
        def log_prior(self, x):
            return -np.inf

    model = BrokenModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert "valid point" in str(excinfo.value)


@pytest.mark.parametrize("log_p", [np.inf, -np.inf])
def test_verify_log_prior_finite(log_p):
    """
    Test `Model.verify_model` and ensure a model with a log-prior that
    only returns inf function raises the correct error
    """

    class BrokenModel(BasicModel):
        def log_prior(self, x):
            return log_p

    model = BrokenModel()

    with pytest.raises(RuntimeError):
        model.verify_model()


def test_verify_log_prior_none():
    """
    Test `Model.verify_model` and ensure a model with a log-prior that
    only returns None raises an error.
    """

    class BrokenModel(BasicModel):
        def log_prior(self, x):
            return None

    model = BrokenModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert "Log-prior" in str(excinfo.value)


def test_verify_log_likelihood_none():
    """
    Test `Model.verify_model` and ensure a model with a log-likelihood that
    only returns None raises an error.
    """

    class BrokenModel(BasicModel):
        def log_likelihood(self, x):
            return None

    model = BrokenModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert "Log-likelihood" in str(excinfo.value)


def test_verify_no_names():
    """
    Test `Model.verify_model` and ensure a model without names
    function raises the correct error
    """

    class TestModel(EmptyModel):
        def __init__(self):
            self.bounds = {"x": [-5, 5], "y": [-5, 5]}

    model = TestModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()
    assert "`names` is not set" in str(excinfo.value)


def test_verify_empty_names():
    """Assert an error is raised if names evaluates to false."""

    class TestModel(EmptyModel):
        names = []

        def __init__(self):
            self.bounds = {"x": [-2, 2], "y": [-2, 2]}

    model = TestModel()
    with pytest.raises(ValueError) as excinfo:
        model.verify_model()

    assert "`names` is not set to a valid value" in str(excinfo.value)


def test_verify_invalid_names_type():
    """Assert an error is raised if names is not a list."""

    class TestModel(EmptyModel):
        names = "x"

        def __init__(self):
            self.bounds = {"x": [-2, 2], "y": [-2, 2]}

    model = TestModel()
    with pytest.raises(TypeError) as excinfo:
        model.verify_model()

    assert "`names` must be a list" in str(excinfo.value)


def test_verify_no_bounds():
    """
    Test `Model.verify_model` and ensure a model without bounds
    function raises the correct error
    """

    class TestModel(EmptyModel):
        def __init__(self):
            self.names = ["x", "y"]

    model = TestModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert "`bounds` is not set" in str(excinfo.value)


def test_verify_empty_bounds():
    """Assert an error is raised if bounds evaluates to false."""

    class TestModel(EmptyModel):
        bounds = {}

        def __init__(self):
            self.names = ["x", "y"]

    model = TestModel()

    with pytest.raises(ValueError) as excinfo:
        model.verify_model()

    assert "`bounds` is not set to a valid value" in str(excinfo.value)


def test_verify_invalid_bounds_type():
    """Assert an error is raised if bounds are not a dictionary."""

    class TestModel(EmptyModel):
        bounds = []

        def __init__(self):
            self.names = ["x", "y"]

    model = TestModel()

    with pytest.raises(TypeError) as excinfo:
        model.verify_model()

    assert "`bounds` must be a dictionary" in str(excinfo.value)


def test_verify_incomplete_bounds():
    """
    Test `Model.verify_model` and ensure a model without bounds
    function raises the correct error
    """

    class TestModel(EmptyModel):
        bounds = {"x": [-5, 5]}

        def __init__(self):
            self.names = ["x", "y"]

    model = TestModel()

    with pytest.raises(RuntimeError):
        model.verify_model()


def test_verify_model_1d():
    """Assert an error is raised if the model is one dimensional."""

    class TestModel(EmptyModel):
        names = ["x"]
        bounds = {"x": [-5, 5]}

    model = TestModel()

    with pytest.raises(OneDimensionalModelError) as excinfo:
        model.verify_model()
    assert "nessai is not designed to handle one-dimensional models" in str(
        excinfo.value
    )


@pytest.mark.parametrize("value", [np.log(True), np.float16(5.0)])
def test_verify_float16(caplog, value):
    """
    Test `Model.verify_model` and ensure that a critical warning is raised
    if a float16 array is returned by the prior.
    """

    class BrokenModel(BasicModel):
        def log_prior(self, x):
            return value

    model = BrokenModel()

    model.verify_model()

    assert "float16 precision" in caplog.text


def test_verify_no_float16(caplog):
    """
    Test `Model.verify_model` and ensure that a critical warning is not raised
    if array return by log_prior is not dtype float16.
    """
    model = BasicModel()
    out = model.verify_model()
    assert out is True
    assert "float16 precision" not in caplog.text


def test_unbounded_priors_wo_new_point():
    """Test `Model.verify_model` with unbounded priors"""

    class TestModel(Model):
        def __init__(self):
            self.names = ["x", "y"]
            self.bounds = {"x": [-5, 5], "y": [-np.inf, np.inf]}

        def log_prior(self, x):
            return -np.log(10) * np.ones(x.size) + norm.logpdf(x["y"])

        def log_likelihood(self, x):
            return np.ones(x.size)

    model = TestModel()
    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert "Could not draw a new point" in str(excinfo.value)


def test_unbounded_priors_w_new_point():
    """Test `Model.verify_model` with unbounded priors"""

    class TestModel(Model):
        def __init__(self):
            self.names = ["x", "y"]
            self.bounds = {"x": [-5, 5], "y": [-np.inf, np.inf]}

        def new_point(self, N=1):
            return numpy_array_to_live_points(
                np.concatenate(
                    [np.random.uniform(-5, 5, (N, 1)), norm.rvs(size=(N, 1))],
                    axis=1,
                ),
                self.names,
            )

        def log_prior(self, x):
            return -np.log(10) * np.ones(x.size) + norm.logpdf(x["y"])

        def log_likelihood(self, x):
            return np.ones(x.size)

    model = TestModel()
    model.verify_model()


def test_verify_model_likelihood_repeated_calls():
    """Assert that an error is raised if repeated calls with the likelihood
    return different values.
    """

    class BrokenModel(BasicModel):
        count = 0
        allow_multi_valued_likelihood = False

        def log_likelihood(self, x):
            self.count += 1
            return self.count

    model = BrokenModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()
    assert "Repeated calls" in str(excinfo.value)


def test_verify_model_likelihood_repeated_calls_allowed(caplog):
    """Assert allow multi valued likelihood prevents an error from being
    raised.
    """

    class MultiValuedModel(BasicModel):
        allow_multi_valued_likelihood = True

        def log_likelihood(self, x):
            return np.random.rand()

    model = MultiValuedModel()
    model.verify_model()
    assert "Multi-valued likelihood is allowed." in str(caplog.text)


def test_configure_pool_with_pool(model):
    """Test configuring the pool when pool is specified"""
    n_pool = 2
    pool = MagicMock()
    with patch("nessai.model.get_n_pool", return_value=n_pool) as mock:
        Model.configure_pool(model, pool=pool, n_pool=1)
    mock.assert_called_once_with(pool)
    assert model.pool is pool
    assert model.n_pool == n_pool


def test_configure_pool_with_pool_no_n_pool(model):
    """Test configuring the pool when pool is specified but n_pool cannot be
    determined and the user has not specified the value.
    """
    pool = MagicMock()
    with patch("nessai.model.get_n_pool", return_value=None) as mock:
        Model.configure_pool(model, pool=pool)
    mock.assert_called_once_with(pool)
    assert model.pool is pool
    assert model.n_pool is None
    assert model.allow_vectorised is False


def test_configure_pool_with_pool_user_n_pool(model):
    """Test configuring the pool when pool is specified but n_pool cannot be
    determined but the user has specified the value.
    """
    model.allow_vectorised = True
    pool = MagicMock()
    with patch("nessai.model.get_n_pool", return_value=None) as mock:
        Model.configure_pool(model, pool=pool, n_pool=1)
    mock.assert_called_once_with(pool)
    assert model.pool is pool
    assert model.n_pool == 1


def test_configure_pool_n_pool(model):
    """Test configuring the pool when n_pool is specified"""
    n_pool = 1
    pool = MagicMock()
    with patch("multiprocessing.Pool", return_value=pool) as mock_pool, patch(
        "nessai.utils.multiprocessing.check_multiprocessing_start_method"
    ) as mock_check:
        Model.configure_pool(model, n_pool=n_pool)
    assert model.pool is pool

    mock_check.assert_called_once()
    mock_pool.assert_called_once_with(
        processes=n_pool,
        initializer=initialise_pool_variables,
        initargs=(model,),
    )


def test_configure_pool_none(model, caplog):
    """Test configuring the pool when pool and n_pool are None"""
    caplog.set_level(logging.INFO)
    Model.configure_pool(model, pool=None, n_pool=None)
    assert model.pool is None
    assert "pool and n_pool are none, no multiprocessing pool" in str(
        caplog.text
    )


def test_configure_pool_already_configured(model, caplog):
    """Assert configuration is skipped if the pool is already configured"""
    model._pool_configured = True
    model.n_pool = 2
    Model.configure_pool(model, n_pool=4)
    assert model.n_pool == 2
    assert "pool has already been configured" in str(caplog.text)


@pytest.mark.parametrize("code", [10, 2])
def test_close_pool(model, code):
    """Test closing the pool"""
    pool = MagicMock()
    pool.close = MagicMock()
    pool.terminate = MagicMock()
    pool.join = MagicMock()
    model.pool = pool
    Model.close_pool(model, code=code)
    pool.join.assert_called_once()
    if code == 2:
        pool.terminate.assert_called_once()
        pool.pool.assert_not_called()
    else:
        pool.close.assert_called_once()
        pool.terminate.assert_not_called()
    assert model.pool is None


@pytest.mark.parametrize("chunksize", [None, 3])
def test_evaluate_likelihoods_pool_vectorised(model, chunksize):
    """Test evaluating a vectorised likelihood with a pool."""
    samples = numpy_array_to_live_points(
        np.array([1, 2, 3, 4])[:, np.newaxis], ["x"]
    )
    logL = [np.array([-1, -2]), np.array([-3, -4])]
    expected = np.array([-1, -2, -3, -4])
    model.pool = MagicMock(side_effect=True)
    model.n_pool = 2
    model.vectorised_likelihood = True
    model.allow_vectorised = True
    model.likelihood_chunksize = chunksize
    model.pool.map = MagicMock(return_value=logL)
    model.likelihood_evaluation_time = datetime.timedelta()
    model.likelihood_evaluations = 100

    out = Model.batch_evaluate_log_likelihood(model, samples)

    assert model.pool.map.call_args_list[0][0][0] is log_likelihood_wrapper
    input_array = model.pool.map.call_args_list[0][0][1]
    if chunksize is None:
        assert len(input_array) == 2
        assert_structured_arrays_equal(
            input_array, np.array([samples[:2], samples[2:]])
        )
    else:
        assert len(input_array) == 2
        for i, a in enumerate(input_array):
            assert_structured_arrays_equal(
                a, samples[i * chunksize : (i + 1) * chunksize]
            )

    model.likelihood_evaluation_time.total_seconds() > 0
    assert model.likelihood_evaluations == 104
    np.testing.assert_array_equal(out, expected)


@pytest.mark.parametrize("chunksize", [None, 1])
def test_evaluate_likelihoods_pool_not_vectorised(model, chunksize):
    """Test evaluating the likelihood with a pool"""
    samples = numpy_array_to_live_points(np.array([[1], [2]]), ["x"])
    logL = np.array([3, 4])
    model.pool = MagicMock(side_effect=True)
    model.n_pool = 2
    model.vectorised_likelihood = False
    model.allow_vectorised = True
    model.likelihood_chunksize = chunksize
    model.pool.map = MagicMock(return_value=logL)
    model.likelihood_evaluation_time = datetime.timedelta()
    model.likelihood_evaluations = 100
    out = Model.batch_evaluate_log_likelihood(model, samples)
    model.pool.map.assert_called_once_with(log_likelihood_wrapper, samples)
    model.likelihood_evaluation_time.total_seconds() > 0
    assert model.likelihood_evaluations == 102
    np.testing.assert_array_equal(out, logL)


def test_evaluate_likelihoods_no_pool_not_vectorised(model):
    """Test evaluating the likelihood without a pool"""
    samples = numpy_array_to_live_points(np.array([[1], [2]]), ["x"])
    # Cannot compare NaNs in has calls
    samples["logL"] = 0.0
    samples["logP"] = 0.0
    logL = np.array([3, 4])
    model.pool = None
    model.n_pool = None
    model.vectorised_likelihood = False
    model.allow_vectorised = True
    model.likelihood_evaluation_time = datetime.timedelta()
    model.likelihood_evaluations = 100
    model.log_likelihood = MagicMock(side_effect=logL)
    out = Model.batch_evaluate_log_likelihood(model, samples)
    model.log_likelihood.assert_has_calls([call(samples[0]), call(samples[1])])
    model.likelihood_evaluation_time.total_seconds() > 0
    assert model.likelihood_evaluations == 102
    np.testing.assert_array_equal(out, logL)


def test_evaluate_likelihoods_no_pool_vectorised(model):
    """
    Test evaluating the likelihood without a pool but with a vectorised
    likelihood.
    """
    samples = numpy_array_to_live_points(np.array([[1], [2]]), ["x"])
    logL = np.array([3, 4])
    model.pool = None
    model.n_pool = None
    model.vectorised_likelihood = True
    model.allow_vectorised = True
    model.likelihood_chunksize = None
    model.likelihood_evaluation_time = datetime.timedelta()
    model.likelihood_evaluations = 100
    model.log_likelihood = MagicMock(return_value=logL)
    out = Model.batch_evaluate_log_likelihood(model, samples)
    model.log_likelihood.assert_called_once_with(samples)
    model.likelihood_evaluation_time.total_seconds() > 0
    assert model.likelihood_evaluations == 102
    np.testing.assert_array_equal(out, logL)


@pytest.mark.parametrize("chunksize", [10, 12])
def test_evaluate_likelihood_vectorised_chunksize(model, chunksize):
    """Assert the likelihood is called the correct number of times"""
    n = 100
    n_calls = np.ceil(n / chunksize)
    samples = numpy_array_to_live_points(np.random.rand(n, 1), ["x"])
    model.vectorised_likelihood = True
    model.allow_vectorised = True
    model.pool = None
    model.n_pool = None
    model.likelihood_chunksize = chunksize
    model.log_likelihood = MagicMock(
        side_effect=lambda x: np.random.rand(x.size)
    )
    out = Model.batch_evaluate_log_likelihood(model, samples)
    assert model.log_likelihood.call_count == n_calls
    assert len(out) == n


def test_evaluate_likelihoods_allow_vectorised_false(model):
    """Assert that vectorisation isn't used if allow_vectorised is false"""
    samples = numpy_array_to_live_points(np.array([[1], [2]]), ["x"])
    logL = [3, 4]
    model.pool = None
    model.n_pool = None
    model.vectorised_likelihood = True
    model.allow_vectorised = False
    model.likelihood_evaluation_time = datetime.timedelta()
    model.likelihood_evaluations = 100
    model.log_likelihood = MagicMock(side_effect=logL)
    out = Model.batch_evaluate_log_likelihood(model, samples)
    assert model.log_likelihood.call_count == 2
    model.likelihood_evaluation_time.total_seconds() > 0
    assert model.likelihood_evaluations == 102
    np.testing.assert_array_equal(out, logL)


def test_evaluate_likelihoods_pool_allow_vectorised_false(model):
    """Test evaluating the likelihood with a pool and allow_vectorised=False"""
    samples = numpy_array_to_live_points(np.array([[1], [2]]), ["x"])
    logL = np.array([3, 4])
    model.pool = MagicMock(side_effect=True)
    model.n_pool = 2
    model.vectorised_likelihood = True
    model.allow_vectorised = False
    model.pool.map = MagicMock(return_value=logL)
    model.likelihood_evaluation_time = datetime.timedelta()
    model.likelihood_evaluations = 100
    out = Model.batch_evaluate_log_likelihood(model, samples)
    model.pool.map.assert_called_once_with(log_likelihood_wrapper, samples)
    model.likelihood_evaluation_time.total_seconds() > 0
    assert model.likelihood_evaluations == 102
    np.testing.assert_array_equal(out, logL)


def test_view_dtype(model):
    """Assert view dtype calls the correct functions with the correct inputs"""
    model._dtype = None
    model.names = ["x", "y"]

    array = np.empty(
        (0,), dtype=[(n, "f8") for n in model.names + ["logL" + "logP"]]
    )
    expected = np.dtype([(n, "f8") for n in model.names])

    with patch(
        "nessai.model._unstructured_view_dtype", return_value=expected
    ) as mock, patch(
        "nessai.model.empty_structured_array", return_value=array
    ) as empty_mock:
        dtype = Model._view_dtype.__get__(model)

    empty_mock.assert_called_once_with(0, model.names)
    mock.assert_called_once_with(array, model.names)
    assert model._dtype == expected
    assert dtype == expected


def test_unstructured_view(model, live_points):
    """Assert the underlying function is called with the correct inputs"""
    out = np.random.randn(live_points.size, 2)
    dtype = np.dtype([("x", "f8"), ("y", "f8")])
    model._view_dtype = dtype
    with patch("nessai.model.unstructured_view", return_value=out) as mock:
        view = Model.unstructured_view(model, live_points)

    mock.assert_called_once_with(live_points, dtype=dtype)
    assert view is out


def test_get_state(model):
    """Assert pool is removed in __getstate__"""
    pool = True
    model.pool = pool
    d = Model.__getstate__(model)
    assert d["pool"] is None
    assert model.pool is pool


@pytest.mark.integration_test
@pytest.mark.parametrize("pickleable", [False, True])
@pytest.mark.parametrize("init", ["before", "after", "function"])
def test_pool(integration_model, mp_context, pickleable, init):
    """Integration test for evaluating the likelihood with a pool"""
    method = mp_context.get_start_method()

    if not pickleable:
        # Cannot pickle lambda functions
        integration_model.fn = lambda x: x
        if method != "fork":
            pytest.xfail(f"start method {method} requires a pickleable model")

    if init == "before":
        if method != "fork":
            pytest.xfail(f"Must use initializer in Pool with {method}")
        initialise_pool_variables(integration_model)
        pool = mp_context.Pool(1)
    elif init == "function":
        pool = mp_context.Pool(
            1,
            initializer=initialise_pool_variables,
            initargs=(integration_model,),
        )
    elif init == "after":
        if method != "fork":
            pytest.xfail(f"Must use initializer in Pool with {method}")
        pool = mp_context.Pool(1)
        initialise_pool_variables(integration_model)
    else:
        raise ValueError(init)

    integration_model.configure_pool(pool=pool)
    assert integration_model.pool is pool
    x = integration_model.new_point(10)
    out = integration_model.batch_evaluate_log_likelihood(x)

    target = np.array(
        [integration_model.log_likelihood(xx) for xx in x],
        dtype=config.livepoints.logl_dtype,
    ).flatten()
    np.testing.assert_array_equal(out, target)
    assert integration_model.likelihood_evaluations == 10

    integration_model.close_pool()
    assert integration_model.pool is None


@pytest.mark.requires("ray")
@pytest.mark.integration_test
@pytest.mark.flaky(reruns=3)
@pytest.mark.skip(reason="Skipping test with ray")
def test_pool_ray(integration_model):
    """Integration test for evaluating the likelihood with a pool from ray.

    This will break if the class for integration_model is defined globally.
    """
    from ray.util.multiprocessing import Pool

    # Cannot pickle lambda functions
    integration_model.fn = lambda x: x
    pool = Pool(
        processes=1,
        initializer=initialise_pool_variables,
        initargs=(integration_model,),
    )
    integration_model.configure_pool(pool=pool)
    assert integration_model.pool is pool
    x = integration_model.new_point(10)
    out = integration_model.batch_evaluate_log_likelihood(x)

    target = np.array(
        [integration_model.log_likelihood(xx) for xx in x],
        dtype=config.livepoints.logl_dtype,
    ).flatten()
    np.testing.assert_array_equal(out, target)
    assert integration_model.likelihood_evaluations == 10

    integration_model.close_pool()
    assert integration_model.pool is None


@pytest.mark.integration_test
@pytest.mark.parametrize("pickleable", [False, True])
def test_n_pool(integration_model, mp_context, pickleable):
    """Integration test for evaluating the likelihood with n_pool"""
    if not pickleable:
        # Cannot pickle lambda functions
        integration_model.fn = lambda x: x
        method = mp_context.get_start_method()
        if method != "fork":
            pytest.xfail(f"start method {method} requires a pickleable model")

    with patch("multiprocessing.Pool", mp_context.Pool), patch(
        "nessai.utils.multiprocessing.multiprocessing.get_start_method",
        mp_context.get_start_method,
    ):
        integration_model.configure_pool(n_pool=1)
    assert integration_model.n_pool == 1
    x = integration_model.new_point(10)
    out = integration_model.batch_evaluate_log_likelihood(x)

    target = np.array(
        [integration_model.log_likelihood(xx) for xx in x],
        dtype=config.livepoints.logl_dtype,
    ).flatten()
    np.testing.assert_array_equal(out, target)
    assert integration_model.likelihood_evaluations == 10

    integration_model.close_pool()
    assert integration_model.pool is None


@pytest.mark.integration_test
def test_unstructured_view_integration(integration_model, live_points):
    """Integration test for unstructured view."""
    view = integration_model.unstructured_view(live_points)
    assert view.base is live_points
    assert view.shape == (live_points.size, integration_model.dims)


@pytest.mark.integration_test
def test_in_bounds_integration_values(integration_model):
    """Assert the correct booleans are returned"""
    x = integration_model.new_point(3)
    names = integration_model.names
    x[names[0]][0] = integration_model.bounds[names[0]][1] + 1.0
    x[names[1]][1] = integration_model.bounds[names[1]][0] - 1.0

    expected = np.array([False, False, True])
    out = integration_model.in_bounds(x)
    np.testing.assert_equal(out, expected)


@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.integration_test
def test_in_bounds_integration_n_samples(integration_model, n):
    """Assert single and multiple samples work"""
    x = numpy_array_to_live_points(
        np.random.randn(n, integration_model.dims), integration_model.names
    )
    flags = integration_model.in_bounds(x)
    assert len(flags) == n



# -*- coding: utf-8 -*-
"""
Functions and objects related to the main nested sampling algorithm.
"""
from collections import deque
from copy import copy
import datetime
import logging
import os
from typing import Union

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import numpy as np
from tqdm import tqdm

from .base import BaseNestedSampler
from .. import config
from ..livepoint import empty_structured_array
from ..plot import plot_indices, plot_trace, nessai_style
from ..evidence import _NSIntegralState
from ..proposal import FlowProposal
from ..proposal.utils import check_proposal_kwargs
from ..utils import (
    compute_indices_ks_test,
    rolling_mean,
)

logger = logging.getLogger(__name__)


class NestedSampler(BaseNestedSampler):
    """
    Nested Sampler class.

    Parameters
    ----------
    model: :obj:`nessai.model.Model`
        User defined model
    nlive : int, optional
        Number of live points.
    output : str
        Path for output
    stopping : float, optional
        Stop when remaining samples wouldn't change logZ estimate by this much.
    max_iteration : int, optional
        Maximum number of iterations to run before force sampler to stop.
        If stopping criteria is met before max. is reached sampler will stop.
    checkpointing : bool, optional
        Boolean to toggle checkpointing, must be enabled to resume the sampler.
        If false the sampler is still saved at the end of sampling.
    checkpoint_interval : int
        The interval used for checkpointing. By default this is a time interval
        in seconds. If :code:`checkpoint_on_iteration=True` this corresponds to
        the number of iterations between checkpointing.
    checkpoint_on_iteration : bool
        If true the checkpointing interval is checked against the number of
        iterations
    logging_interval : int, optional
        The interval in seconds used for periodic logging. If not specified,
        then periodic logging is disabled.
    log_on_iteration : bool
        If true logging will occur based on the iteration. If false logging
        will be periodic if `logging_interval` is set. In case where neither
        logging is enabled, `log_on_iteration` will be set to true with an
        interval of :code:`nlive`.
    resume_file : str, optional
        If specified sampler will be resumed from this file. Still requires
        correct model.
    seed : int, optional
        seed for the initialisation of the pseudorandom chain
    n_pool : int, optional
        Number of threads to when for creating the multiprocessing pool.
    pool : object
        User defined multiprocessing pool that will be used when evaluating
        the likelihood.
    close_pool : bool
        Boolean to indicated if the pool should be closed at the end of the
        nested sampling loop. If False, the user must manually close the pool.
    plot : bool (True)
        Boolean to toggle plotting
    proposal_plots : bool (True)
        Boolean to enable additional plots for the population stage of the
        sampler. Overwritten by plot.
    prior_sampling : bool (False)
        produce nlive samples from the prior.
    analytic_priors : bool (False)
        Boolean that indicates that the `new_point` method in the model
        draws directly from the priors meaning rejection sampling is not
        needed.
    maximum_uninformed : int (1000)
        Maximum number of iterations before forcing the sampler to switch to
        using the proposal method with the flow.
    uninformed_proposal : :obj:`nessai.proposal.Proposal`: (None)
        Class to use for initial sampling before training the flow. If
        None RejectionProposal or AnalyticProposal are used depending if
        `analytic_priors` is False or True.
    uninformed_acceptance_threshold : float (None)
        Acceptance threshold for initialising sampling, if acceptance falls
        below this value sampler switches to flow-based proposal. If None
        then value is set to 10 times `acceptance_threshold`
    uninformed_proposal_kwargs : dict, ({})
        Dictionary of keyword argument to pass to the class use for
        the initial sampling when it is initialised.
    flow_class : :obj:`nessai.proposal.FlowProposal`
        Class to use for flow-based proposal method
    flow_config : dict ({})
        Dictionary used to configure instance of `nessai.flowmodel.FlowModel`,
        this includes configuring the normalising flow and the training.
    training_frequency : int (None)
        Number of iterations between re-training the flow. If None flow
        is only re-trained based on other criteria.
    train_on_empty : bool (True)
        If true the flow is retrained every time the proposal pool is
        empty. If false it is only training according to the other criteria.
    cooldown : int (100)
        Minimum number of iterations between training. Can be overridden if
        `train_on_empty=True` and the pool is empty.
    memory : int, False (False)
        Number of old live points to use in training. If False only the current
        live points are used.
    reset_weights : bool, int, (False)
        Boolean to toggle resetting the flow weights whenever re-training.
        If an integer is specified the flow is reset every nth time it is
        trained.
    reset_permutations: bool, int, (False)
        Boolean to toggle resetting the permutation layers in the flow whenever
        re-training. If an integer is specified the flow is reset every nth
        time it is trained.
    reset_flow : bool, int
        Boolean to toggle resetting the entire flow whenever re-training. If
        an integer is specified the flow is reset every nth time it trains.
        This option overrides :code:`reset_weights` and
        :code:`reset_permutations`.
    reset_acceptance : bool, (True)
        If true use mean acceptance of samples produced with current flow
        as a criteria for retraining
    retrain_acceptance : bool (False)
        Force the flow to be reset if the acceptance falls below the acceptance
        threshold. Requires `reset_acceptance=True`
    acceptance_threshold : float (0.01)
        Threshold to determine if the flow should be retrained, will not
        retrain if cooldown is not satisfied.
    shrinkage_expectation : str, {"t", "logt"}
        Method used to compute the expectation value for the shrinkage t.
        Choose between log <t> or <log t>. Defaults to <log t>.
    trace_parameters : Optional[list[str]]
        List of parameters to include in the trace plot. If None, all model
        parameters are included.
    kwargs :
        Keyword arguments passed to the flow proposal class
    """

    def __init__(
        self,
        model,
        nlive=2000,
        output=None,
        stopping=0.1,
        max_iteration=None,
        checkpointing=True,
        checkpoint_interval=600,
        checkpoint_on_iteration=False,
        checkpoint_on_training=False,
        checkpoint_callback=None,
        logging_interval=None,
        log_on_iteration=True,
        resume_file=None,
        seed=None,
        pool=None,
        close_pool=False,
        n_pool=None,
        plot=True,
        proposal_plots=False,
        prior_sampling=False,
        analytic_priors=False,
        maximum_uninformed=None,
        uninformed_proposal=None,
        uninformed_acceptance_threshold=None,
        uninformed_proposal_kwargs=None,
        flow_class=None,
        flow_config=None,
        training_frequency=None,
        train_on_empty=True,
        cooldown=200,
        memory=False,
        reset_weights=False,
        reset_permutations=False,
        reset_flow=False,
        retrain_acceptance=True,
        reset_acceptance=False,
        acceptance_threshold=0.01,
        shrinkage_expectation="logt",
        trace_parameters=None,
        **kwargs,
    ):

        super().__init__(
            model,
            nlive,
            output=output,
            seed=seed,
            checkpointing=checkpointing,
            checkpoint_interval=checkpoint_interval,
            checkpoint_on_iteration=checkpoint_on_iteration,
            checkpoint_callback=checkpoint_callback,
            logging_interval=logging_interval,
            log_on_iteration=log_on_iteration,
            resume_file=resume_file,
            plot=plot,
            n_pool=n_pool,
            pool=pool,
        )

        self._close_pool = close_pool

        self.live_points = None
        self.prior_sampling = prior_sampling
        self.accepted = 0
        self.rejected = 1
        self.initialised = False

        self.checkpoint_on_training = checkpoint_on_training
        self.iteration = 0
        self.acceptance_history = deque(maxlen=(nlive // 10))
        self.block_acceptance = 1.0
        self.mean_block_acceptance = 1.0
        self.block_iteration = 0
        self.retrain_acceptance = retrain_acceptance
        self.reset_acceptance = reset_acceptance
        self.trace_parameters = (
            trace_parameters if trace_parameters is not None else model.names
        )

        self.insertion_indices = []
        self.rolling_p = []
        self.final_p_value = None
        self.final_ks_statistic = None

        self.resumed = False
        self.tolerance = stopping
        self.condition = np.inf
        self.logLmin = -np.inf
        self.logLmax = -np.inf
        self.nested_samples = []
        self.logZ = None
        self.state = _NSIntegralState(
            self.nlive,
            track_gradients=plot,
            expectation=shrinkage_expectation,
        )

        # Timing
        self.training_time = datetime.timedelta()

        # Resume flags
        self.completed_training = True

        self.acceptance_threshold = acceptance_threshold

        self.train_on_empty = train_on_empty
        self.cooldown = cooldown
        self.memory = memory

        self.configure_max_iteration(max_iteration)
        self.configure_flow_reset(
            reset_weights, reset_permutations, reset_flow
        )
        self.configure_training_frequency(training_frequency)

        if uninformed_proposal_kwargs is None:
            uninformed_proposal_kwargs = {}
        self.configure_uninformed_proposal(
            uninformed_proposal,
            analytic_priors,
            maximum_uninformed,
            uninformed_acceptance_threshold,
            **uninformed_proposal_kwargs,
        )
        self.configure_flow_proposal(
            flow_class, flow_config, proposal_plots, **kwargs
        )

        # Uninformed proposal is used for prior sampling
        # If maximum uninformed is greater than 0, the it will be used for
        # another n iterations or until it becomes inefficient

    @property
    def log_evidence(self):
        return self.state.logZ

    @property
    def log_evidence_error(self):
        return self.state.log_evidence_error

    @property
    def information(self):
        return self.state.info[-1]

    @property
    def proposal_population_time(self):
        t = self._uninformed_proposal.population_time
        t += self._flow_proposal.population_time
        return t

    @property
    def acceptance(self):
        return self.iteration / self.likelihood_calls

    @property
    def last_updated(self):
        """Last time the normalising flow was retrained"""
        if self.history and self.history["training_iterations"]:
            return self.history["training_iterations"][-1]
        else:
            return 0

    @property
    def mean_acceptance(self):
        """
        Mean acceptance of the last nlive // 10 points
        """
        if self.acceptance_history:
            return np.mean(self.acceptance_history)
        else:
            return np.nan

    @property
    def posterior_effective_sample_size(self):
        """The effective sample size of the posterior distribution"""
        return self.state.effective_n_posterior_samples

    @property
    def birth_log_likelihoods(self):
        """Return the birth log-likelihoods for the nested samples"""
        logLs = np.array(self.state.logLs)
        its = np.array(self.nested_samples)["it"]
        return logLs[its].flatten()

    def configure_max_iteration(self, max_iteration):
        """Configure the maximum iteration.

        If None then no maximum is set.

        Parameters
        ----------
        max_iteration : int, None
            Maximum iteration.
        """
        if max_iteration is None:
            self.max_iteration = np.inf
        else:
            self.max_iteration = max_iteration

    def configure_training_frequency(self, training_frequency):
        """Configure the training frequency.

        If None, 'inf' or 'None' flow will only train when empty.
        """
        if training_frequency in [None, "inf", "None"]:
            logger.debug("Proposal will only train when empty")
            self.training_frequency = np.inf
        else:
            self.training_frequency = training_frequency

    def configure_uninformed_proposal(
        self,
        uninformed_proposal,
        analytic_priors,
        maximum_uninformed,
        uninformed_acceptance_threshold,
        **kwargs,
    ):
        """
        Setup the uninformed proposal method (is NOT trained)

        Parameters
        ----------
        uninformed_proposal : None or obj
            Class to use for uninformed proposal
        analytic_priors : bool
            If True `AnalyticProposal` is used to directly sample from the
            priors rather than using rejection sampling.
        maximum_uninformed : {False, None, int, float}
            Maximum number of iterations before switching to FlowProposal.
            If None, two times nlive is used. If False uninformed sampling is
            not used.
        uninformed_acceptance_threshold :  float or None:
            Threshold to use for uninformed proposal, once reached proposal
            method will switch. If None acceptance_threshold is used if
            greater than 0.1 else 10 x acceptance_threshold is used.
        kwargs
            Kwargs are passed to init method for uninformed proposal class
        """
        if maximum_uninformed is None:
            self.uninformed_sampling = True
            self.maximum_uninformed = 2 * self.nlive
        elif not maximum_uninformed:
            self.uninformed_sampling = False
            self.maximum_uninformed = 0
        else:
            self.uninformed_sampling = True
            self.maximum_uninformed = float(maximum_uninformed)

        if uninformed_acceptance_threshold is None:
            if self.acceptance_threshold < 0.1:
                self.uninformed_acceptance_threshold = (
                    10 * self.acceptance_threshold
                )
            else:
                self.uninformed_acceptance_threshold = (
                    self.acceptance_threshold
                )
        else:
            self.uninformed_acceptance_threshold = (
                uninformed_acceptance_threshold
            )

        if uninformed_proposal is None:
            if analytic_priors:
                from ..proposal import AnalyticProposal as uninformed_proposal
            else:
                from ..proposal import RejectionProposal as uninformed_proposal

                kwargs["poolsize"] = self.nlive

        logger.debug(f"Using uninformed proposal: {uninformed_proposal}")
        logger.debug(f"Parsing kwargs to uninformed proposal: {kwargs}")
        self._uninformed_proposal = uninformed_proposal(self.model, **kwargs)

    def configure_flow_proposal(
        self, flow_class, flow_config, proposal_plots, **kwargs
    ):
        """
        Set up the flow-based proposal method

        Parameters
        ----------
        flow_class : None or obj or str
            Class to use for proposal. If None FlowProposal is used.
        flow_config : dict
            Configuration dictionary passed to the class.
        proposal_plots : bool or str
            Configuration of plotting in proposal class.
        **kwargs :
            Kwargs passed to init function.
        """
        proposal_output = os.path.join(self.output, "proposal", "")

        if not self.plot:
            proposal_plots = False

        if flow_class is not None:
            if isinstance(flow_class, str):
                flow_class = flow_class.lower()
                if flow_class == "gwflowproposal":
                    from ..gw.proposal import GWFlowProposal as flow_class
                elif flow_class == "augmentedgwflowproposal":
                    from ..gw.proposal import (
                        AugmentedGWFlowProposal as flow_class,
                    )
                elif flow_class == "flowproposal":
                    flow_class = FlowProposal
                elif flow_class == "augmentedflowproposal":
                    from ..proposal import AugmentedFlowProposal

                    flow_class = AugmentedFlowProposal
                else:
                    raise ValueError(f"Unknown flow class: {flow_class}")
            elif not issubclass(flow_class, FlowProposal):
                raise RuntimeError(
                    "Flow class must be string or class that "
                    "inherits from FlowProposal"
                )
        else:
            flow_class = FlowProposal

        if kwargs.get("poolsize", None) is None:
            kwargs["poolsize"] = self.nlive

        logger.debug(f"Using flow class: {flow_class}")

        kwargs = check_proposal_kwargs(flow_class, kwargs)

        logger.info(f"Passing kwargs to {flow_class.__name__}: {kwargs}")
        self._flow_proposal = flow_class(
            self.model,
            flow_config=flow_config,
            output=proposal_output,
            plot=proposal_plots,
            **kwargs,
        )

    def configure_output(
        self, output: Union[str, None], resume_file: Union[str, None] = None
    ):
        """Configure the output folder

        Creates a "diagnostics" directory plotting is enabled.

        Parameters
        ----------
        output : str
            Directory where the results will be stored
        resume_file : str, optional
            Specific file to use for checkpointing. If not specified the
            default is used (nested_sampler_resume.pkl)
        """
        super().configure_output(output, resume_file)
        if self.plot:
            os.makedirs(os.path.join(output, "diagnostics"), exist_ok=True)

    def configure_flow_reset(
        self, reset_weights, reset_permutations, reset_flow
    ):
        """Configure how often the flow parameters are reset.

        Values are converted to floats.

        Parameters
        ----------
        reset_weights : int, float or bool
            Frequency with which the weights will be reset.
        reset_permutations : int, float or bool
            Frequency with which the permutations will be reset.
        reset_flow : int, float or bool
            Frequency with which the entire flow will be reset. Will overwrite
            the value for the other resets.
        """
        if isinstance(reset_weights, (int, float)):
            self.reset_weights = float(reset_weights)
        else:
            raise TypeError("`reset_weights` must be a bool, int or float")
        if isinstance(reset_permutations, (int, float)):
            self.reset_permutations = float(reset_permutations)
        else:
            raise TypeError(
                "`reset_permutations` must be a bool, int or float"
            )
        if isinstance(reset_flow, (int, float)):
            self.reset_flow = float(reset_flow)
        else:
            raise TypeError("`reset_flow` must be a bool, int or float")

        if self.reset_flow:
            self.reset_weights = self.reset_flow
            self.reset_permutations = self.reset_flow

    def initialise_history(self):
        if not self.history:
            super().initialise_history()
            self.history.update(
                dict(
                    iterations=[],
                    min_log_likelihood=[],
                    max_log_likelihood=[],
                    logZ=[],
                    dlogZ=[],
                    mean_acceptance=[],
                    training_iterations=[],
                    population_acceptance=[],
                    population_radii=[],
                    population_iterations=[],
                    rolling_p=[],
                )
            )

    def update_history(self):
        super().update_history()
        self.history["iterations"].append(self.iteration)
        self.history["min_log_likelihood"].append(self.logLmin)
        self.history["max_log_likelihood"].append(self.logLmax)
        self.history["logZ"].append(self.state.logZ)
        self.history["dlogZ"].append(self.condition)
        self.history["mean_acceptance"].append(self.mean_acceptance)

    def log_state(self):
        """Log the current state of the sampler"""
        logger.info(
            f"it: {self.iteration:5d}: "
            f"n eval: {self.likelihood_calls} "
            f"H: {self.state.info[-1]:.2f} "
            f"dlogZ: {self.condition:.3f} logZ: {self.state.logZ:.3f} "
            f"+/- {self.state.log_evidence_error:.3f} "
            f"logLmax: {self.logLmax:.2f}"
        )

    def check_insertion_indices(self, rolling=True, filename=None):
        """
        Checking the distribution of the insertion indices either during
        the nested sampling run (rolling=True) or for the whole run
        (rolling=False).
        """
        if rolling:
            indices = self.insertion_indices[-self.nlive :]
        else:
            indices = self.insertion_indices

        D, p = compute_indices_ks_test(indices, self.nlive)

        if p is not None:
            if rolling:
                logger.info(
                    f"it: {self.iteration:5d}: "
                    f"Rolling KS test: D={D:.4}, p-value={p:.4}"
                )
                self.history["rolling_p"].append(p)
            else:
                logger.info(f"Final KS test: D={D:.4}, p-value={p:.4}")
                self.final_p_value = p
                self.final_ks_statistic = D
                pt = 0.05
                if self.final_p_value < pt:
                    logger.warning(
                        f"Final p-value for the insertion indices is less "
                        f"than {pt}, this could be an indication of problems "
                        "during sampling. Consider checking the diagnostic "
                        "plots."
                    )

        if filename is not None:
            np.savetxt(
                os.path.join(self.output, filename),
                self.insertion_indices,
                newline="\n",
                delimiter=" ",
            )

    def yield_sample(self, oldparam):
        """
        Draw points and applying rejection sampling
        """
        while True:
            counter = 0
            while True:
                counter += 1
                newparam = self.proposal.draw(copy(oldparam))

                # Prior is computed in the proposal
                if newparam["logP"] != -np.inf:
                    if not newparam["logL"]:
                        newparam["logL"] = self.model.evaluate_log_likelihood(
                            newparam
                        )
                    if newparam["logL"] > self.logLmin:
                        self.logLmax = max(self.logLmax, newparam["logL"])
                        oldparam = newparam.copy()
                        break
                # Only here if proposed and then empty
                # This returns the old point and allows for a training check
                if not self.proposal.populated:
                    break
            yield counter, oldparam

    def insert_live_point(self, live_point):
        """
        Insert a live point
        """
        # This is the index including the current worst point, so final index
        # is one less, otherwise index=0 would never be possible
        index = np.searchsorted(self.live_points["logL"], live_point["logL"])
        self.live_points[: index - 1] = self.live_points[1:index]
        self.live_points[index - 1] = live_point
        return index - 1

    def consume_sample(self):
        """
        Replace a sample for single thread
        """
        worst = self.live_points[0].copy()
        self.logLmin = worst["logL"]
        self.state.increment(worst["logL"])
        self.nested_samples.append(worst)

        self.condition = (
            np.logaddexp(
                self.state.logZ,
                self.logLmax - self.iteration / float(self.nlive),
            )
            - self.state.logZ
        )

        # Replace the points we just consumed with the next acceptable ones
        # Make sure we are mixing the chains
        self.iteration += 1
        self.block_iteration += 1
        count = 0

        while True:
            c, proposed = next(self.yield_sample(worst))
            count += c

            if proposed["logL"] > self.logLmin:
                # Assuming point was proposed
                # replace worst point with new one
                proposed["it"] = self.iteration
                index = self.insert_live_point(proposed)
                self.insertion_indices.append(index)
                self.accepted += 1
                self.block_acceptance += 1 / count
                self.acceptance_history.append(1 / count)
                break
            else:
                # Only get here if the yield sample returns worse point
                # which can only happen if the pool is empty
                self.rejected += 1
                self.check_state()
                # if retrained whilst proposing a sample then update the
                # iteration count since will be zero otherwise
                if not self.block_iteration:
                    self.block_iteration += 1

        self.mean_block_acceptance = (
            self.block_acceptance / self.block_iteration
        )

        if self.debug_enabled and self.log_on_iteration:
            logger.debug(
                f"{self.iteration:5d}: n: {count:3d} "
                f"b_acc: {self.mean_block_acceptance:.3f} "
                f"H: {self.state.info[-1]:.2f} "
                f"logL: {self.logLmin:.5f} --> {proposed['logL']:.5f} "
                f"dlogZ: {self.condition:.3f} "
                f"logZ: {self.state.logZ:.3f} "
                f"+/- {self.state.log_evidence_error:.3f} "
                f"logLmax: {self.logLmax:.2f}"
            )

    def populate_live_points(self):
        """
        Initialise the pool of live points.
        """
        i = 0
        live_points = empty_structured_array(
            self.nlive, names=self.model.names
        )

        with tqdm(total=self.nlive, desc="Drawing live points") as pbar:
            while i < self.nlive:
                while i < self.nlive:
                    _, live_point = next(self.yield_sample(None))
                    if live_point is None:
                        break
                    if np.isnan(live_point["logL"]):
                        logger.error(
                            "Likelihood function returned NaN for "
                            f"live_point {live_point}"
                        )
                        logger.error(
                            "You may want to check your likelihood function"
                        )
                        break
                    if np.isfinite(live_point["logP"]) and np.isfinite(
                        live_point["logL"]
                    ):
                        live_points[i] = live_point
                        i += 1
                        pbar.update()
                        break

        self.live_points = np.sort(live_points, order="logL")
        self.live_points["it"] = 0

    def initialise(self, live_points=True):
        """
        Initialise the nested sampler

        Parameters
        ----------
        live_points : bool, optional (True)
            If true and there are no live points, new live points are
            drawn using `populate_live_points` else all other initialisation
            steps are complete but live points remain empty.
        """
        flags = [False] * 3
        if not self._flow_proposal.initialised:
            self._flow_proposal.initialise()
            flags[0] = True

        if not self._uninformed_proposal.initialised:
            self._uninformed_proposal.initialise()
            flags[1] = True

        if (
            self.iteration < self.maximum_uninformed
            and self.uninformed_sampling
        ):
            self.proposal = self._uninformed_proposal
        else:
            self.proposal = self._flow_proposal

        if live_points and self.live_points is None:
            self.populate_live_points()
            flags[2] = True

        if self.condition > self.tolerance:
            self.finalised = False

        self.initialise_history()

        if all(flags):
            self.initialised = True

    def check_proposal_switch(self, force=False):
        """
        Check if the proposal should be switch from uninformed to
        flowproposal given the current state.

        If the flow proposal is already in use, no changes are made.

        Parameters
        ----------
        force : bool, optional
            If True proposal is forced to switch.

        Returns
        -------
        bool
            Flag to indicated if proposal was switched
        """
        if (
            (self.mean_acceptance < self.uninformed_acceptance_threshold)
            or (self.iteration >= self.maximum_uninformed)
            or force
        ):
            if self.proposal is self._flow_proposal:
                logger.info("Already using flowproposal")
                self.uninformed_sampling = False
                return True
            logger.info("Switching to FlowProposal")
            self.proposal = self._flow_proposal
            self.proposal.ns_acceptance = self.mean_block_acceptance
            self.uninformed_sampling = False
            return True
        # If using uninformed sampling, don't check training
        else:
            return False

    def check_training(self):
        """
        Check if the normalising flow should be trained

        Checks that can force training:
            - Training was previously stopped before completion
            - The pool is empty and the proposal was not in the process
              of populating when stopped.
        Checks that cannot force training is still on cooldown:
            - Acceptance falls below threshold and `retrain_acceptance` is
              true
            - The number of iterations since last training is equal to the
              training frequency

        Returns
        -------
        train : bool
            Try to train if true
        force : bool
            Force the training irrespective of cooldown
        """
        if not self.completed_training:
            logger.debug("Training flow (resume)")
            return True, True
        elif (
            not self.proposal.populated
            and self.train_on_empty
            and not self.proposal.populating
        ):
            logger.debug("Training flow (proposal empty)")
            return True, True
        elif (
            self.mean_block_acceptance < self.acceptance_threshold
            and self.retrain_acceptance
        ):
            logger.debug("Training flow (acceptance)")
            return True, False
        elif (self.iteration - self.last_updated) == self.training_frequency:
            logger.debug("Training flow (iteration)")
            return True, False
        else:
            return False, False

    def check_flow_model_reset(self):
        """
        Check if the normalising flow model should be reset.

        Checks acceptance if `reset_acceptance` is True and always checks
        how many times the flow has been trained.

        Flow will not be reset if it has not been trained. To force a reset
        manually call `proposal.reset_model_weights`.
        """
        if not self.proposal.training_count:
            return

        if (
            self.reset_acceptance
            and self.mean_block_acceptance < self.acceptance_threshold
        ):
            self.proposal.reset_model_weights(weights=True, permutations=True)
            return

        self.proposal.reset_model_weights(
            weights=(
                self.reset_weights
                and not (self.proposal.training_count % self.reset_weights)
            ),
            permutations=(
                self.reset_permutations
                and not (
                    self.proposal.training_count % self.reset_permutations
                )
            ),
        )

    def train_proposal(self, force=False):
        """
        Try to train the proposal. Proposal will not train if cooldown is not
        exceeded unless force is True.

        Parameters
        ----------
        force : bool
            Override training checks
        """
        if self.iteration - self.last_updated < self.cooldown and not force:
            logger.debug("Not training, still cooling down!")
        else:
            self.completed_training = False
            self.check_flow_model_reset()

            training_data = self.live_points.copy()
            if self.memory and (len(self.nested_samples) >= self.memory):
                training_data = np.concatenate(
                    [training_data, self.nested_samples[-self.memory :].copy()]
                )

            st = datetime.datetime.now()
            self.proposal.train(training_data)
            self.training_time += datetime.datetime.now() - st
            self.history["training_iterations"].append(self.iteration)

            self.block_iteration = 0
            self.block_acceptance = 0.0
            self.completed_training = True
            if self.checkpoint_on_training:
                self.checkpoint(periodic=True)

    def check_state(self, force=False):
        """
        Check if state should be updated prior to drawing a new sample

        Force will override the cooldown mechanism.
        """

        if self.uninformed_sampling:
            if self.check_proposal_switch():
                force = True
            else:
                return
        # General override
        train = False
        if force:
            train = True
            logger.debug("Training flow (force)")
        elif not train:
            train, force = self.check_training()

        if train or force:
            self.train_proposal(force=force)

    @nessai_style()
    def plot_state(self, filename=None):
        """
        Produce plots with the current state of the nested sampling run.
        Plots are saved to the output directory specified at initialisation.

        Parameters
        ----------
        filename : str, optional
            If specified the figure will be saved, otherwise the figure is
            returned.
        """

        fig, ax = plt.subplots(7, 1, sharex=True, figsize=(12, 12))
        ax = ax.ravel()
        it = np.array(self.history["iterations"])

        for i in self.history["checkpoint_iterations"]:
            for a in ax:
                a.axvline(i, ls=":", color="#66ccff")

        for a in ax:
            a.axvline(self.iteration, c="#ff9900", ls="-.")

        ax[0].plot(it, self.history["min_log_likelihood"], label="Min log L")
        ax[0].plot(it, self.history["max_log_likelihood"], label="Max log L")
        ax[0].set_ylabel(r"$\log L$")
        ax[0].legend(frameon=False)

        logX_its = np.arange(len(self.state.log_vols))
        ax[1].plot(logX_its, self.state.log_vols, label="log X")
        ax[1].set_ylabel(r"$\log X$")
        ax[1].legend(frameon=False)

        if self.state.track_gradients:
            ax_logX_grad = plt.twinx(ax[1])
            # Use dotted linestyle (ls[2]) because dashed isn't clear
            ax_logX_grad.plot(
                logX_its,
                rolling_mean(np.abs(self.state.gradients), self.nlive // 10),
                c="C1",
                ls=config.plotting.line_styles[1],
                label="Gradient",
            )
            ax_logX_grad.set_ylabel(r"$|d\log L/d \log X|$")
            ax_logX_grad.set_yscale("log")
            handles, labels = ax[1].get_legend_handles_labels()
            handles_tw, labels_tw = ax_logX_grad.get_legend_handles_labels()
            ax[1].legend(
                handles + handles_tw, labels + labels_tw, frameon=False
            )

        ax[2].plot(
            it, self.history["likelihood_evaluations"], label="Evaluations"
        )
        ax[2].set_ylabel("Likelihood\n evaluations")
        ax[2].set_yscale("log")
        ax[2].legend()

        ax_time = plt.twinx(ax[2])
        ax_time.plot(
            it,
            np.array(self.history["sampling_time"]) / 60,
            ls=config.plotting.line_styles[1],
            color="C1",
            label="Time",
        )
        ax_time.set_ylabel("Sampling \ntime [min]")
        handles, labels = ax[2].get_legend_handles_labels()
        handles_time, labels_time = ax_time.get_legend_handles_labels()
        ax[2].legend(
            handles + handles_time, labels + labels_time, frameon=False
        )

        ax[3].plot(it, self.history["logZ"], label="logZ")
        ax[3].set_ylabel(r"$\log Z$")
        ax[3].legend(frameon=False)

        ax_dz = plt.twinx(ax[3])
        ax_dz.plot(
            it,
            self.history["dlogZ"],
            label="dlogZ",
            c="C1",
            ls=config.plotting.line_styles[1],
        )
        ax_dz.set_yscale("log")
        ax_dz.set_ylabel(r"$d \log Z$")
        handles, labels = ax[3].get_legend_handles_labels()
        handles_dz, labels_dz = ax_dz.get_legend_handles_labels()
        ax[3].legend(handles + handles_dz, labels + labels_dz, frameon=False)

        ax[4].plot(it, self.history["mean_acceptance"], label="Proposal")
        ax[4].plot(
            self.history["population_iterations"],
            self.history["population_acceptance"],
            label="Population",
        )
        ax[4].set_ylabel("Acceptance")
        handles, labels = ax[4].get_legend_handles_labels()

        ax_r = plt.twinx(ax[4])
        ax_r.plot(
            self.history["population_iterations"],
            self.history["population_radii"],
            label="Radius",
            color="C2",
            ls=config.plotting.line_styles[2],
        )
        ax_r.set_ylabel("Population radius")
        handles_r, labels_r = ax_r.get_legend_handles_labels()
        ax[4].legend(handles + handles_r, labels + labels_r, frameon=False)
        ax[4].set_yscale("log")
        ax[4].set_ylim(top=1.1)
        dtrain = np.array(self.history["training_iterations"][1:]) - np.array(
            self.history["training_iterations"][:-1]
        )
        ax[5].plot(self.history["training_iterations"][1:], dtrain)
        if self.history["training_iterations"]:
            ax[5].axvline(
                self.history["training_iterations"][0],
                ls="-",
                color="lightgrey",
            )
        ax[5].set_ylabel(r"$\Delta$ train")

        if len(self.history["rolling_p"]):
            it = (np.arange(len(self.history["rolling_p"])) + 1) * self.nlive
            ax[6].plot(it, self.history["rolling_p"], "o", label="p-value")
        ax[6].set_ylabel("p-value")
        ax[6].set_ylim([-0.1, 1.1])

        ax[-1].set_xlabel("Iteration")

        fig.suptitle(
            f"Sampling time: {self.current_sampling_time}", fontsize=16
        )

        handles = [
            Line2D(
                [0],
                [0],
                color="#ff9900",
                linestyle="-.",
                label="Current iteration",
            ),
            Line2D(
                [0], [0], color="#66ccff", linestyle=":", label="Checkpoint"
            ),
        ]
        fig.legend(handles=handles, frameon=False, ncol=3, loc=(0.6, 0.0))

        fig.tight_layout()
        fig.subplots_adjust(top=0.95)
        if filename is not None:
            fig.savefig(filename)
            plt.close(fig)
        else:
            return fig

    def plot_trace(self, filename=None, **kwargs):
        """
        Make trace plots for the nested samples.

        Parameters
        ----------
        filename : str, optional
            If filename is None, the figure is returned. Else the figure
            is saved with that file name.
        kwargs :
            Additional keyword arguments passed to
            :py:func:`nessai.plot.plot_trace`.
        """
        if self.nested_samples:
            if self.live_points is not None:
                log_x_live_points = self.state.get_logx_live_points(self.nlive)
            else:
                log_x_live_points = None
            fig = plot_trace(
                self.state.log_vols[1:],
                self.nested_samples,
                parameters=self.trace_parameters,
                live_points=self.live_points,
                log_x_live_points=log_x_live_points,
                filename=filename,
                **kwargs,
            )
            return fig
        else:
            logger.warning("Could not produce trace plot. No nested samples!")

    def plot_insertion_indices(self, filename=None, **kwargs):
        """
        Make a plot of all the insertion indices.

        Parameters
        ----------
        filename : str, optional
            If filename is None, the figure is returned. Else the figure
            is saved with that file name.
        kwargs :
            Keyword arguments passed to `nessai.plot.plot_indices`.
        """
        return plot_indices(
            self.insertion_indices, self.nlive, filename=filename, **kwargs
        )

    def update_state(self, force=False):
        """
        Update state after replacing a live point
        """
        # Check if acceptance is not None, this indicates the proposal
        # was populated
        if not self.proposal._checked_population:
            self.history["population_acceptance"].append(
                self.proposal.population_acceptance
            )
            self.history["population_radii"].append(self.proposal.r)
            self.history["population_iterations"].append(self.iteration)
            self.proposal._checked_population = True

        if not (self.iteration % (self.nlive // 10)) or force:
            self.update_history()

        if not (self.iteration % self.nlive) or force:
            if not force:
                self.check_insertion_indices()
                if self.plot:
                    plot_indices(
                        self.insertion_indices[-self.nlive :],
                        self.nlive,
                        plot_breakdown=False,
                        filename=os.path.join(
                            self.output,
                            "diagnostics",
                            f"insertion_indices_{self.iteration}.png",
                        ),
                    )

            if self.plot:
                self.plot_state(
                    filename=os.path.join(self.output, "state.png")
                )
                self.plot_trace(
                    filename=os.path.join(self.output, "trace.png")
                )

            if self.uninformed_sampling:
                self.block_acceptance = 0.0
                self.block_iteration = 0

        if self.checkpointing:
            self.checkpoint(periodic=True)

        self.proposal.ns_acceptance = self.mean_block_acceptance

    def check_resume(self):
        """
        Check the normalising flow is correctly configured is the sampler
        was resumed.
        """
        if self.resumed:
            if self.uninformed_sampling is False:
                self.check_proposal_switch(force=True)
            # If pool is populated reset the flag since it is set to
            # false during initialisation
            if hasattr(self._flow_proposal, "resume_populated"):
                if (
                    self._flow_proposal.resume_populated
                    and self._flow_proposal.indices
                ):
                    self._flow_proposal.populated = True
                    logger.info("Resumed with populated pool")

            self.resumed = False

    def finalise(self):
        """
        Finalise things after sampling
        """
        logger.info("Finalising")
        for i, p in enumerate(self.live_points):
            self.state.increment(p["logL"], nlive=self.nlive - i)
            self.nested_samples.append(p)
        self.live_points = None

        # Refine evidence estimate
        self.update_state(force=True)
        self.state.finalise()
        # output the chain and evidence
        self.finalised = True

    def nested_sampling_loop(self):
        """Main nested sampling loop.

        Notes
        -----
        If :code:`prior_sampling=True` then sampling will stop after drawing
        the initial live points.

        Returns
        -------
        log_evidence : float
            The final log evidence.
        nested_samples : numpy.ndarray
            Array of nested samples.
        """
        self.sampling_start_time = datetime.datetime.now()
        if not self.initialised:
            self.initialise(live_points=True)

        if self.prior_sampling:
            if self._close_pool:
                self.close_pool()
            self.finalise()
            return self.log_evidence, np.array(self.nested_samples)

        self.check_resume()

        if self.iteration:
            self.update_state()

        logger.info("Starting nested sampling loop")

        while self.condition > self.tolerance:

            self.check_state()

            self.consume_sample()

            self.update_state()

            self.periodically_log_state()

            if self.iteration >= self.max_iteration:
                logger.info("Reached max iteration")
                break
        logger.info(f"Stopping with dlogZ={self.condition:.3f}")

        # final adjustments
        # avoid repeating final adjustments if resuming a completed run.
        if not self.finalised and (self.condition <= self.tolerance):
            self.finalise()

        logger.info(
            f"Final ln-evidence: {self.state.logZ:.3f} +/- "
            f"{self.state.log_evidence_error:.3f}"
        )
        logger.info(f"Information: {self.state.info[-1]:.2f}")

        self.check_insertion_indices(rolling=False)

        # This includes updating the total sampling time
        self.checkpoint(periodic=True, force=True)

        if self._close_pool:
            self.close_pool()

        logger.info(f"Total sampling time: {self.sampling_time}")
        logger.info(f"Total training time: {self.training_time}")
        logger.info(f"Total population time: {self.proposal_population_time}")
        logger.info(
            f"Total likelihood evaluations: {self.likelihood_calls:3d}"
        )
        logger.info(
            "Time spent evaluating likelihood: "
            f"{self.likelihood_evaluation_time}"
        )

        return self.state.logZ, np.array(self.nested_samples)

    def get_result_dictionary(self):
        """Return a dictionary that contains results"""
        d = super().get_result_dictionary()
        d["insertion_indices"] = self.insertion_indices
        d["final_p_value"] = self.final_p_value
        d["final_ks_statistic"] = self.final_ks_statistic
        d["nested_samples"] = np.array(self.nested_samples)
        d["log_posterior_weights"] = self.state.log_posterior_weights
        d["logL_birth"] = self.birth_log_likelihoods
        d["log_evidence"] = self.log_evidence
        d["log_evidence_error"] = self.state.log_evidence_error
        d["information"] = self.information
        d["training_time"] = self.training_time.total_seconds()
        d["population_time"] = self.proposal_population_time.total_seconds()
        return d

    @classmethod
    def resume_from_pickled_sampler(
        cls, sampler, model, flow_config=None, weights_path=None, **kwargs
    ):
        """Resume from a pickled sampler.

        Parameters
        ----------
        sampler: Any
            Pickled sampler.
        model : :obj:`nessai.model.Model`
            User-defined model
        flow_config : Optional[dict]
            Dictionary for configuring the flow
        weights_path : Optional[str]
            Weights file to use in place of the weights file stored in the
            pickle file.
        kwargs :
            Keyword arguments passed to the parent class's method.

        Returns
        -------
        Instance of NestedSampler
        """
        obj = super(NestedSampler, cls).resume_from_pickled_sampler(
            sampler, model, **kwargs
        )
        obj._uninformed_proposal.resume(model)
        if flow_config is None:
            flow_config = {}
        obj._flow_proposal.resume(model, flow_config, weights_path)
        return obj



from nessai.samplers.importancesampler import OrderedSamples
from nessai.evidence import _INSIntegralState
from nessai.utils.testing import assert_structured_arrays_equal
import numpy as np
import pytest
from unittest.mock import MagicMock, create_autospec, patch


@pytest.fixture()
def ordered_samples():
    return create_autospec(OrderedSamples)


@pytest.fixture()
def samples(samples):
    return np.sort(samples, order="logL")


@pytest.fixture()
def live_points(samples):
    return samples[samples.size // 2 :]


@pytest.fixture()
def nested_samples(samples):
    return samples[: samples.size // 2]


def test_live_points(ordered_samples, samples):
    indices = [2, 3]
    ordered_samples.live_points_indices = indices
    ordered_samples.samples = samples
    out = OrderedSamples.live_points.__get__(ordered_samples)
    assert_structured_arrays_equal(out, samples[indices])


def test_live_points_none(ordered_samples, samples):
    ordered_samples.live_points_indices = None
    ordered_samples.samples = samples
    assert OrderedSamples.live_points.__get__(ordered_samples) is None


def test_live_points_setter_error(ordered_samples):
    with pytest.raises(ValueError, match=r"Can only set live points to None"):
        OrderedSamples.live_points.__set__(ordered_samples, 1.0)


def test_live_points_setter(ordered_samples):
    OrderedSamples.live_points.__set__(ordered_samples, None)
    assert ordered_samples.live_points_indices is None


def test_nested_samples(ordered_samples, samples):
    indices = [2, 3]
    ordered_samples.nested_samples_indices = indices
    ordered_samples.samples = samples
    out = OrderedSamples.nested_samples.__get__(ordered_samples)
    assert_structured_arrays_equal(out, samples[indices])


def test_nested_samples_none(ordered_samples, samples):
    ordered_samples.nested_samples_indices = None
    ordered_samples.samples = samples
    assert OrderedSamples.nested_samples.__get__(ordered_samples) is None


def test_update_log_likelihood_threshold(ordered_samples):
    val = 5.0
    ordered_samples.log_likelihood_threshold = None
    OrderedSamples.update_log_likelihood_threshold(ordered_samples, val)
    assert ordered_samples.log_likelihood_threshold == val


def test_sort_samples_only(ordered_samples):
    samples = np.array(np.random.randn(10), [("logL", "f8")])
    samples_out = OrderedSamples.sort_samples(ordered_samples, samples)
    assert np.all(np.diff(samples_out["logL"]) > 0)


def test_sort_samples_with_log_q(ordered_samples):
    samples = np.array(np.random.randn(10), [("logL", "f8")])
    order = np.argsort(samples["logL"])
    extra = np.arange(samples.size)
    sorted_samples, sorted_extra = OrderedSamples.sort_samples(
        ordered_samples, samples, extra
    )
    assert_structured_arrays_equal(sorted_samples, samples[order])
    np.testing.assert_array_equal(sorted_extra, extra[order])


def test_add_initial_samples(ordered_samples, samples, log_q):
    ordered_samples.sort_samples = MagicMock(return_value=(samples, log_q))
    OrderedSamples.add_initial_samples(ordered_samples, samples, log_q)
    ordered_samples.sort_samples.assert_called_once_with(samples, log_q)
    assert ordered_samples.samples is samples
    assert ordered_samples.log_q is log_q
    assert len(ordered_samples.live_points_indices) == len(samples)
    assert np.all(np.diff(ordered_samples.live_points_indices) > 0)
    np.testing.assert_array_equal(
        ordered_samples.live_points_indices, np.arange(samples.size)
    )


@pytest.mark.parametrize("has_live_points", [True, False])
def test_add_samples_soft(ordered_samples, samples, log_q, has_live_points):
    n = int(0.8 * samples.size)
    ordered_samples.strict_threshold = False

    if has_live_points:
        n_ns = int(0.8 * n)
        ns_indices = np.sort(np.random.choice(n, size=n_ns, replace=False))
        live_indices = np.sort(list(set(np.arange(n)) - set(ns_indices)))
    else:
        n_ns = n
        ns_indices = np.arange(n_ns)
        live_indices = None

    sort_idx = np.argsort(samples[:n], order="logL")
    ordered_samples.samples = samples[:n][sort_idx]
    ordered_samples.log_q = log_q[:n][sort_idx]
    ordered_samples.nested_samples_indices = ns_indices
    ordered_samples.live_points_indices = live_indices

    new_samples = samples[n:]
    new_log_q = log_q[n:]

    idx = np.argsort(new_samples, order="logL")
    new_samples_ordered = new_samples[idx]
    new_log_q_ordered = new_log_q[idx]

    ordered_samples.sort_samples = MagicMock(
        return_value=(new_samples_ordered, new_log_q_ordered)
    )

    OrderedSamples.add_samples(ordered_samples, new_samples, new_log_q)

    ordered_samples.sort_samples.assert_called_once_with(
        new_samples, new_log_q
    )
    assert len(ordered_samples.live_points_indices) == (
        n - n_ns + new_samples.size
    )
    assert np.all(np.diff(ordered_samples.samples["logL"]) >= 0)
    assert np.all(
        np.diff(
            ordered_samples.samples[ordered_samples.live_points_indices][
                "logL"
            ]
        )
        >= 0
    )
    assert np.all(
        np.diff(
            ordered_samples.samples[ordered_samples.nested_samples_indices][
                "logL"
            ]
        )
        >= 0
    )


def test_add_samples(ordered_samples, samples, log_q):
    ordered_samples.strict_threshold = True

    idx = np.argsort(samples, order="logL")
    expected = samples[idx].copy()
    expected_log_q = log_q[idx].copy()

    perm = np.random.permutation(samples.size)
    samples = samples[perm]
    log_q = log_q[perm]
    n = int(0.8 * samples.size)
    idx = np.argsort(samples[:n], order="logL")
    ordered_samples.samples = samples[:n][idx]
    ordered_samples.log_q = log_q[:n][idx]

    new_samples = samples[n:]
    new_log_q = log_q[n:]

    idx = np.argsort(new_samples, order="logL")
    new_samples_ordered = new_samples[idx]
    new_log_q_ordered = new_log_q[idx]
    ordered_samples.sort_samples = MagicMock(
        return_value=(new_samples_ordered, new_log_q_ordered)
    )

    ordered_samples.log_likelihood_threshold = new_samples_ordered[
        new_samples_ordered.size // 2
    ]["logL"].item()

    n_expected = np.sum(
        expected["logL"] >= ordered_samples.log_likelihood_threshold
    )

    OrderedSamples.add_samples(ordered_samples, new_samples, new_log_q)

    assert_structured_arrays_equal(ordered_samples.samples, expected)
    np.testing.assert_array_equal(ordered_samples.log_q, expected_log_q)
    np.testing.assert_array_equal(
        ordered_samples.nested_samples_indices,
        np.arange(samples.size - n_expected),
    )
    np.testing.assert_array_equal(
        ordered_samples.live_points_indices,
        np.arange(samples.size - n_expected, samples.size),
    )


@pytest.mark.parametrize("replace_all", [False, True])
def test_remove_samples(ordered_samples, replace_all):
    n = 10
    ordered_samples.live_points = np.array(
        np.arange(n), dtype=[("logL", "f8")]
    )
    ordered_samples.live_points_indices = np.arange(n)
    ordered_samples.log_likelihood_threshold = 5.5
    ordered_samples.replace_all = replace_all
    ordered_samples.add_to_nested_samples = MagicMock()

    expected = n if replace_all else 6

    out = OrderedSamples.remove_samples(ordered_samples)

    assert out == expected
    ordered_samples.add_to_nested_samples.assert_called_once()
    np.testing.assert_array_equal(
        ordered_samples.add_to_nested_samples.call_args.args[0],
        np.arange(expected),
    )
    if replace_all:
        assert ordered_samples.live_points_indices is None
    else:
        np.testing.assert_array_equal(
            ordered_samples.live_points_indices, np.arange(6, n)
        )


def test_add_to_nested_samples(ordered_samples):
    ns_indices = np.array([0, 1, 2, 4, 5, 8])
    indices = np.array([3, 6, 7, 9])
    ordered_samples.nested_samples_indices = ns_indices
    OrderedSamples.add_to_nested_samples(ordered_samples, indices)
    np.testing.assert_array_equal(
        ordered_samples.nested_samples_indices, np.arange(10)
    )


def test_update_evidence(ordered_samples, live_points, nested_samples):
    ordered_samples.state = create_autospec(_INSIntegralState)
    ordered_samples.live_points = live_points
    ordered_samples.nested_samples = nested_samples
    OrderedSamples.update_evidence(ordered_samples)
    ordered_samples.state.update_evidence.assert_called_once_with(
        nested_samples=nested_samples,
        live_points=live_points,
    )


def test_finalise(ordered_samples, samples):
    ordered_samples.samples = samples
    ordered_samples.live_points_indices = np.arange(4)

    def add_to_ns(indices):
        ordered_samples.live_points_indices = None

    ordered_samples.add_to_nested_samples = MagicMock(side_effect=add_to_ns)
    ordered_samples.state = create_autospec(_INSIntegralState)

    OrderedSamples.finalise(ordered_samples)

    ordered_samples.state.update_evidence.assert_called_once_with(samples)
    assert ordered_samples.live_points is None
    assert ordered_samples.live_points_indices is None


@pytest.mark.parametrize("ratio", [0.0, 0.5, 1.0])
def test_compute_importance(ordered_samples, log_q, samples, ratio):
    ordered_samples.samples = samples
    ordered_samples.log_q = log_q
    out = OrderedSamples.compute_importance(
        ordered_samples, importance_ratio=ratio
    )
    assert len(set(out.keys()) - {"total", "posterior", "evidence"}) == 0
    assert np.all(np.isfinite(list(out.values())))


@pytest.mark.parametrize("threshold", [None, 0])
def test_computed_evidence_ratio(ordered_samples, samples, threshold):
    log_z_total = -10.0
    log_z = -6.0
    ordered_samples.samples = samples
    ordered_samples.log_likelihood_threshold = np.median(samples["logL"])
    ordered_samples.state = MagicMock(spec=_INSIntegralState)
    ordered_samples.state.log_evidence = log_z_total

    with patch(
        "nessai.samplers.importancesampler.log_evidence_from_ins_samples",
        return_value=log_z,
    ) as mock_log_evidence:
        out = OrderedSamples.compute_evidence_ratio(ordered_samples, threshold)
    actual_threshold = (
        ordered_samples.log_likelihood_threshold
        if threshold is None
        else threshold
    )
    above_threshold = samples["logL"] > actual_threshold

    mock_log_evidence.assert_called_once()
    assert_structured_arrays_equal(
        mock_log_evidence.call_args_list[0][0][0], samples[above_threshold]
    )
    assert out == (log_z - log_z_total)


@pytest.mark.parametrize("save_log_q", [False, True])
def test_getstate(ordered_samples, save_log_q):
    samples = np.random.randn(20, 4)
    log_q = np.random.randn(2, 20)
    ordered_samples.save_log_q = save_log_q
    ordered_samples.log_q = log_q
    ordered_samples.samples = samples
    state = OrderedSamples.__getstate__(ordered_samples)
    assert state["samples"] is samples
    if save_log_q:
        assert state["log_q"] is log_q
    else:
        assert state["log_q"] is None



# -*- coding: utf-8 -*-
"""
Tests for utilities related to insertion indices.
"""
import numpy as np
import pytest

from nessai.utils.indices import compute_indices_ks_test, bonferroni_correction


@pytest.mark.parametrize("mode", ["D+", "D-"])
def test_ks_test(mode):
    """
    Test KS test for insertion indices with a specified mode
    """
    indices = np.random.randint(0, 1000, 1000)
    out = compute_indices_ks_test(indices, 1000, mode=mode)
    assert all([o > 0.0 for o in out])


def test_ks_test_undefined_mode():
    """
    Test KS test for insertion indices with undefined mode
    """
    indices = np.random.randint(0, 1000, 1000)
    with pytest.raises(RuntimeError):
        compute_indices_ks_test(indices, 1000, mode="two-sided")


def test_ks_test_empty_indices():
    """
    Test KS test for insertion indices with empty input array
    """
    out = compute_indices_ks_test([], 1000, mode="D+")
    assert all(o is None for o in out)


def test_bonferroni_correction():
    """Test the Bonferroni correction for p-values"""
    p_values = np.linspace(0, 0.5, 4)
    rejected, corrected, alpha = bonferroni_correction(p_values)
    np.testing.assert_array_equal(corrected, np.array([0, 2 / 3, 1, 1]))
    assert rejected.tolist() == [True, False, False, False]
    assert alpha == 0.0125
